"""Structures for the eQ-3 Bluetooth Smart Thermostat."""

from dataclasses import dataclass
from datetime import datetime, time, timedelta
from typing import Self

from construct import (
    Bytes,
    Const,
    Flag,
    GreedyBytes,
    GreedyRange,
    Int8ub,
    Optional,
)
from construct_typed import DataclassMixin, DataclassStruct, TEnum, TFlagsEnum, csfield

from eq3btsmart._adapters import (
    _Eq3AwayTime,
    _Eq3Duration,
    _Eq3ScheduleTime,
    _Eq3Serial,
    _Eq3Temperature,
    _Eq3TemperatureOffset,
    _Eq3Time,
)
from eq3btsmart.const import (
    Eq3WeekDay,
    _Eq3Command,
    _Eq3StatusFlags,
)

__all__: list[str] = []


class _Eq3Struct(DataclassMixin):
    """Structure for eQ-3 data."""

    @classmethod
    def from_bytes(cls, data: bytes) -> Self:
        """Convert the data to a structure."""
        return DataclassStruct(cls).parse(data)

    def to_bytes(self) -> bytes:
        """Convert the structure to bytes."""
        return DataclassStruct(self.__class__).build(self)


@dataclass
class _DeviceDataStruct(_Eq3Struct):
    """Structure for device data."""

    cmd: int = csfield(Const(_Eq3Command.ID_RETURN, Int8ub))
    version: int = csfield(Int8ub)
    unknown_1: int = csfield(Int8ub)
    unknown_2: int = csfield(Int8ub)
    serial: str = csfield(_Eq3Serial(Bytes(10)))
    unknown_3: int = csfield(Int8ub)


@dataclass
class _PresetsStruct(_Eq3Struct):
    """Structure for presets data."""

    window_open_temp: float = csfield(_Eq3Temperature(Int8ub))
    window_open_time: timedelta = csfield(_Eq3Duration(Int8ub))
    comfort_temp: float = csfield(_Eq3Temperature(Int8ub))
    eco_temp: float = csfield(_Eq3Temperature(Int8ub))
    offset: float = csfield(_Eq3TemperatureOffset(Int8ub))


@dataclass
class _StatusStruct(_Eq3Struct):
    """Structure for status data."""

    cmd: int = csfield(Const(_Eq3Command.INFO_RETURN, Int8ub))
    const_1: int = csfield(Const(0x01, Int8ub))
    mode: _Eq3StatusFlags = csfield(TFlagsEnum(Int8ub, _Eq3StatusFlags))
    valve: int = csfield(Int8ub)
    const_2: int = csfield(Const(0x04, Int8ub))
    target_temp: float = csfield(_Eq3Temperature(Int8ub))
    away: datetime | None = csfield(Optional(_Eq3AwayTime(Bytes(4))))
    presets: _PresetsStruct | None = csfield(Optional(DataclassStruct(_PresetsStruct)))


@dataclass
class _ScheduleHourStruct(_Eq3Struct):
    """Structure for schedule entry data."""

    target_temp: float = csfield(_Eq3Temperature(Int8ub))
    next_change_at: time = csfield(_Eq3ScheduleTime(Int8ub))


@dataclass
class _ScheduleDayStruct(_Eq3Struct):
    """Structure for schedule data."""

    cmd: int = csfield(Const(_Eq3Command.SCHEDULE_RETURN, Int8ub))
    day: Eq3WeekDay = csfield(TEnum(Int8ub, Eq3WeekDay))
    hours: list[_ScheduleHourStruct] = csfield(
        GreedyRange(DataclassStruct(_ScheduleHourStruct))
    )


@dataclass
class _Eq3Message(_Eq3Struct):
    """Structure for eQ-3 message."""

    cmd: int = csfield(Int8ub)
    is_status_command: bool = csfield(Flag)
    data: bytes = csfield(GreedyBytes)


@dataclass
class _IdGetCommand(_Eq3Struct):
    """Structure for ID get command."""

    cmd: int = csfield(Const(_Eq3Command.ID_GET, Int8ub))


@dataclass
class _InfoGetCommand(_Eq3Struct):
    """Structure for info get command."""

    cmd: int = csfield(Const(_Eq3Command.INFO_GET, Int8ub))
    time: datetime = csfield(_Eq3Time(Bytes(6)))


@dataclass
class _ComfortEcoConfigureCommand(_Eq3Struct):
    """Structure for schedule get command."""

    cmd: int = csfield(Const(_Eq3Command.COMFORT_ECO_CONFIGURE, Int8ub))
    comfort_temperature: float = csfield(_Eq3Temperature(Int8ub))
    eco_temperature: float = csfield(_Eq3Temperature(Int8ub))


@dataclass
class _OffsetConfigureCommand(_Eq3Struct):
    """Structure for offset configure command."""

    cmd: int = csfield(Const(_Eq3Command.OFFSET_CONFIGURE, Int8ub))
    offset: float = csfield(_Eq3TemperatureOffset(Int8ub))


@dataclass
class _WindowOpenConfigureCommand(_Eq3Struct):
    """Structure for window open configure command."""

    cmd: int = csfield(Const(_Eq3Command.WINDOW_OPEN_CONFIGURE, Int8ub))
    window_open_temperature: float = csfield(_Eq3Temperature(Int8ub))
    window_open_time: timedelta = csfield(_Eq3Duration(Int8ub))


@dataclass
class _ScheduleGetCommand(_Eq3Struct):
    """Structure for schedule get command."""

    cmd: int = csfield(Const(_Eq3Command.SCHEDULE_GET, Int8ub))
    day: Eq3WeekDay = csfield(TEnum(Int8ub, Eq3WeekDay))


@dataclass
class _ModeSetCommand(_Eq3Struct):
    """Structure for mode set command."""

    cmd: int = csfield(Const(_Eq3Command.MODE_SET, Int8ub))
    mode: int = csfield(Int8ub)


@dataclass
class _AwaySetCommand(_ModeSetCommand):
    """Structure for away set command."""

    away_until: datetime = csfield(_Eq3AwayTime(Bytes(4)))


@dataclass
class _TemperatureSetCommand(_Eq3Struct):
    """Structure for temperature set command."""

    cmd: int = csfield(Const(_Eq3Command.TEMPERATURE_SET, Int8ub))
    temperature: float = csfield(_Eq3Temperature(Int8ub))


@dataclass
class _ScheduleSetCommand(_Eq3Struct):
    """Structure for schedule set command."""

    cmd: int = csfield(Const(_Eq3Command.SCHEDULE_SET, Int8ub))
    day: Eq3WeekDay = csfield(TEnum(Int8ub, Eq3WeekDay))
    hours: list[_ScheduleHourStruct] = csfield(
        GreedyRange(DataclassStruct(_ScheduleHourStruct))
    )


@dataclass
class _ComfortSetCommand(_Eq3Struct):
    """Structure for comfort set command."""

    cmd: int = csfield(Const(_Eq3Command.COMFORT_SET, Int8ub))


@dataclass
class _EcoSetCommand(_Eq3Struct):
    """Structure for eco set command."""

    cmd: int = csfield(Const(_Eq3Command.ECO_SET, Int8ub))


@dataclass
class _BoostSetCommand(_Eq3Struct):
    """Structure for boost set command."""

    cmd: int = csfield(Const(_Eq3Command.BOOST_SET, Int8ub))
    enable: bool = csfield(Flag)


@dataclass
class _LockSetCommand(_Eq3Struct):
    """Structure for lock set command."""

    cmd: int = csfield(Const(_Eq3Command.LOCK_SET, Int8ub))
    enable: bool = csfield(Flag)



from unittest.mock import MagicMock

import pytest
from construct import Construct

from eq3btsmart._adapters import _Eq3TemperatureOffset
from eq3btsmart.const import EQ3_MAX_OFFSET, EQ3_MIN_OFFSET
from eq3btsmart.exceptions import Eq3InvalidDataException


def test_encode_valid_temperature_offset() -> None:
    offset = 1.0
    encoded = _Eq3TemperatureOffset.encode(offset)
    assert encoded == 9


def test_decode_valid_int() -> None:
    encoded_int = 9
    decoded = _Eq3TemperatureOffset.decode(encoded_int)
    assert decoded == 1.0


def test_round_trip_consistency() -> None:
    offset = 1.0
    encoded = _Eq3TemperatureOffset.encode(offset)
    decoded = _Eq3TemperatureOffset.decode(encoded)
    assert decoded == offset


def test_encode_temperature_offset_out_of_range_low() -> None:
    offset = EQ3_MIN_OFFSET - 0.1
    with pytest.raises(Eq3InvalidDataException):
        _Eq3TemperatureOffset.encode(offset)


def test_encode_temperature_offset_out_of_range_high() -> None:
    offset = EQ3_MAX_OFFSET + 0.1
    with pytest.raises(Eq3InvalidDataException):
        _Eq3TemperatureOffset.encode(offset)


def test_encode() -> None:
    subcon = MagicMock(spec=Construct)
    subcon.flagbuildnone = False
    offset = 1.0
    value = _Eq3TemperatureOffset(subcon)
    encoded = value._encode(offset, MagicMock(), "")
    assert encoded == 9


def test_decode() -> None:
    subcon = MagicMock(spec=Construct)
    subcon.flagbuildnone = False
    encoded = 9
    value = _Eq3TemperatureOffset(subcon)
    decoded = value._decode(encoded, MagicMock(), "")
    assert decoded == 1.0



"""Models for the eQ-3 Bluetooth Smart Thermostat library."""

from __future__ import annotations

from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from datetime import datetime, time, timedelta
from typing import Self

from eq3btsmart._structures import (
    _DeviceDataStruct,
    _Eq3Struct,
    _PresetsStruct,
    _ScheduleDayStruct,
    _StatusStruct,
)
from eq3btsmart.const import EQ3_OFF_TEMP, EQ3_ON_TEMP, Eq3OperationMode, Eq3WeekDay

__all__ = [
    "DeviceData",
    "Presets",
    "Schedule",
    "ScheduleDay",
    "ScheduleHour",
    "Status",
]


@dataclass
class _BaseModel[StructType: _Eq3Struct](ABC):
    @classmethod
    @abstractmethod
    def _from_struct(cls: type[Self], struct: StructType) -> Self:
        """Convert the structure to a model."""

    @classmethod
    @abstractmethod
    def _struct_type(cls: type[Self]) -> type[StructType]:
        """Return the structure type associated with the model."""

    @classmethod
    def _from_bytes(cls: type[Self], data: bytes) -> Self:
        """Convert the data to a model."""
        return cls._from_struct(cls._struct_type().from_bytes(data))


@dataclass
class DeviceData(_BaseModel[_DeviceDataStruct]):
    """Device data model.

    Contains general information about the device that is affected by commands scheduled by the user.

    Attributes:
        firmware_version(int): The firmware version of the device.
        device_serial(str): The serial number of the device.
    """

    firmware_version: int
    device_serial: str

    @classmethod
    def _from_struct(cls, struct: _DeviceDataStruct) -> Self:
        return cls(
            firmware_version=struct.version,
            device_serial=struct.serial,
        )

    @classmethod
    def _struct_type(cls) -> type[_DeviceDataStruct]:
        return _DeviceDataStruct


@dataclass
class Status(_BaseModel[_StatusStruct]):
    """Status model.

    Contains the current status of the device.

    Attributes:
        valve(int): The current valve position.
        target_temperature(float): The target temperature.
        operation_mode(Eq3OperationMode): The operation mode.
        is_away(bool): Whether the device is in away mode.
        is_boost(bool): Whether the device is in boost mode.
        is_dst(bool): Whether the device is in daylight saving time.
        is_window_open(bool): Whether the window is open.
        is_locked(bool): Whether the device is locked.
        is_low_battery(bool): Whether the battery is low.
        away_until(datetime | None): The time until the device is in away mode.
        presets(Presets | None): The presets of the device.
    """

    valve: int
    target_temperature: float
    _operation_mode: Eq3OperationMode
    is_away: bool
    is_boost: bool
    is_dst: bool
    is_window_open: bool
    is_locked: bool
    is_low_battery: bool
    away_until: datetime | None = None
    presets: Presets | None = None

    @property
    def operation_mode(self) -> Eq3OperationMode:
        """The operation mode."""
        if self.target_temperature == EQ3_OFF_TEMP:
            return Eq3OperationMode.OFF

        if self.target_temperature == EQ3_ON_TEMP:
            return Eq3OperationMode.ON

        return self._operation_mode

    @property
    def valve_temperature(self) -> float:
        """The valve temperature.

        The valve temperature is calculated based on the valve position and the target temperature.
        """
        return (1 - self.valve / 100) * 2 + self.target_temperature - 2

    @classmethod
    def _from_struct(cls, struct: _StatusStruct) -> Self:
        return cls(
            valve=struct.valve,
            target_temperature=struct.target_temp,
            _operation_mode=Eq3OperationMode.MANUAL
            if struct.mode & struct.mode.MANUAL
            else Eq3OperationMode.AUTO,
            is_away=bool(struct.mode & struct.mode.AWAY),
            is_boost=bool(struct.mode & struct.mode.BOOST),
            is_dst=bool(struct.mode & struct.mode.DST),
            is_window_open=bool(struct.mode & struct.mode.WINDOW),
            is_locked=bool(struct.mode & struct.mode.LOCKED),
            is_low_battery=bool(struct.mode & struct.mode.LOW_BATTERY),
            away_until=struct.away,
            presets=Presets._from_struct(struct.presets) if struct.presets else None,
        )

    @classmethod
    def _struct_type(cls) -> type[_StatusStruct]:
        return _StatusStruct


@dataclass
class Presets(_BaseModel[_PresetsStruct]):
    """Presets model.

    Contains the presets of the device.

    Attributes:
        window_open_temperature(float): The temperature when the window is open.
        window_open_time(timedelta): The time the window is open.
        comfort_temperature(float): The comfort temperature.
        eco_temperature(float): The eco temperature.
        offset_temperature(float): The offset temperature.
    """

    window_open_temperature: float
    window_open_time: timedelta
    comfort_temperature: float
    eco_temperature: float
    offset_temperature: float

    @classmethod
    def _from_struct(cls, struct: _PresetsStruct) -> Self:
        return cls(
            window_open_temperature=struct.window_open_temp,
            window_open_time=struct.window_open_time,
            comfort_temperature=struct.comfort_temp,
            eco_temperature=struct.eco_temp,
            offset_temperature=struct.offset,
        )

    @classmethod
    def _struct_type(cls) -> type[_PresetsStruct]:
        return _PresetsStruct


@dataclass
class Schedule:
    """Schedule model.

    The schedule is a list of days with a list of hours for each day.
    Each hour contains the target temperature and the time when the next change occurs.

    Attributes:
        schedule_days(list[ScheduleDay]): The schedule days.
    """

    schedule_days: list[ScheduleDay] = field(default_factory=list)

    def merge(self, other_schedule: Self) -> None:
        """Merge another schedule into this schedule.

        Args:
            other_schedule(Schedule): The other schedule to merge.
        """
        for other_schedule_day in other_schedule.schedule_days:
            schedule_day = next(
                (
                    schedule_day
                    for schedule_day in self.schedule_days
                    if schedule_day.week_day == other_schedule_day.week_day
                ),
                None,
            )

            if not schedule_day:
                self.schedule_days.append(other_schedule_day)
                continue

            schedule_day.schedule_hours = other_schedule_day.schedule_hours

    @classmethod
    def _from_bytes(cls, data: bytes) -> Self:
        return cls(schedule_days=[ScheduleDay._from_bytes(data)])

    def __eq__(self, __value: object) -> bool:
        """Check if the schedule is equal to another schedule.

        Args:
            __value(object): The value to compare.
        """
        if not isinstance(__value, Schedule):
            return False

        week_days_to_compare = [
            schedule_day.week_day
            for schedule_day in self.schedule_days
            if len(schedule_day.schedule_hours) > 0
        ]
        week_days_to_compare.extend(
            [
                schedule_day.week_day
                for schedule_day in __value.schedule_days
                if len(schedule_day.schedule_hours) > 0
            ]
        )

        for week_day in week_days_to_compare:
            schedule_day = next(
                (
                    schedule_day
                    for schedule_day in self.schedule_days
                    if schedule_day.week_day == week_day
                ),
                None,
            )
            other_schedule_day = next(
                (
                    schedule_day
                    for schedule_day in __value.schedule_days
                    if schedule_day.week_day == week_day
                ),
                None,
            )

            if schedule_day is None or other_schedule_day is None:
                return False

            other_schedule_day = next(
                (
                    other_schedule_day
                    for other_schedule_day in __value.schedule_days
                    if other_schedule_day.week_day == schedule_day.week_day
                ),
                None,
            )

            if schedule_day != other_schedule_day:
                return False

        return True


@dataclass
class ScheduleDay(_BaseModel[_ScheduleDayStruct]):
    """Schedule day model.

    Attributes:
        week_day(Eq3WeekDay): The week day.
        schedule_hours(list[ScheduleHour]): The schedule hours.
    """

    week_day: Eq3WeekDay
    schedule_hours: list[ScheduleHour] = field(default_factory=list)

    @classmethod
    def _from_struct(cls, struct: _ScheduleDayStruct) -> Self:
        return cls(
            week_day=struct.day,
            schedule_hours=[
                ScheduleHour(
                    target_temperature=hour.target_temp,
                    next_change_at=hour.next_change_at,
                )
                for hour in struct.hours
            ],
        )

    @classmethod
    def _struct_type(cls) -> type[_ScheduleDayStruct]:
        return _ScheduleDayStruct

    def __eq__(self, __value: object) -> bool:
        """Check if the schedule day is equal to another schedule day.

        Args:
            __value(object): The value to compare.
        """
        if not isinstance(__value, ScheduleDay):
            return False

        if self.week_day != __value.week_day:
            return False

        if len(self.schedule_hours) != len(__value.schedule_hours):
            return False

        return all(
            hour == other_hour
            for hour, other_hour in zip(self.schedule_hours, __value.schedule_hours)
        )


@dataclass
class ScheduleHour:
    """Schedule hour model.

    Attributes:
        target_temperature(float): The target temperature.
        next_change_at(time): The time when the next change occurs.
    """

    target_temperature: float
    next_change_at: time



from unittest.mock import MagicMock

import pytest
from construct import Construct

from eq3btsmart._adapters import _Eq3Temperature
from eq3btsmart.const import EQ3_OFF_TEMP, EQ3_ON_TEMP
from eq3btsmart.exceptions import Eq3InvalidDataException


def test_encode_valid_temperature() -> None:
    temperature = 22.0
    encoded = _Eq3Temperature.encode(temperature)
    assert encoded == 44


def test_decode_valid_int() -> None:
    encoded_int = 44
    decoded = _Eq3Temperature.decode(encoded_int)
    assert decoded == 22.0


def test_round_trip_consistency() -> None:
    temperature = 22.0
    encoded = _Eq3Temperature.encode(temperature)
    decoded = _Eq3Temperature.decode(encoded)
    assert decoded == temperature


def test_encode_temperature_out_of_range_low() -> None:
    temperature = EQ3_OFF_TEMP - 0.1
    with pytest.raises(Eq3InvalidDataException):
        _Eq3Temperature.encode(temperature)


def test_encode_temperature_out_of_range_high() -> None:
    temperature = EQ3_ON_TEMP + 0.1
    with pytest.raises(Eq3InvalidDataException):
        _Eq3Temperature.encode(temperature)


def test_encode() -> None:
    subcon = MagicMock(spec=Construct)
    subcon.flagbuildnone = False
    temperature = 22.0
    value = _Eq3Temperature(subcon)
    encoded = value._encode(temperature, MagicMock(), "")
    assert encoded == 44


def test_decode() -> None:
    subcon = MagicMock(spec=Construct)
    subcon.flagbuildnone = False
    encoded = 44
    value = _Eq3Temperature(subcon)
    decoded = value._decode(encoded, MagicMock(), "")
    assert decoded == 22.0



from unittest.mock import MagicMock

from construct import Construct

from eq3btsmart._adapters import _Eq3Serial


def test_encode_valid_string() -> None:
    string = "example"
    encoded = _Eq3Serial.encode(string)
    assert encoded == bytes([ord(char) + 0x30 for char in string])


def test_decode_valid_bytes() -> None:
    encoded_bytes = bytes([ord(char) + 0x30 for char in "example"])
    decoded = _Eq3Serial.decode(encoded_bytes)
    assert decoded == "example"


def test_round_trip_consistency() -> None:
    string = "example"
    encoded = _Eq3Serial.encode(string)
    decoded = _Eq3Serial.decode(encoded)
    assert decoded == string


def test_encode_empty_string() -> None:
    string = ""
    encoded = _Eq3Serial.encode(string)
    assert encoded == b""


def test_decode_empty_bytes() -> None:
    encoded_bytes = b""
    decoded = _Eq3Serial.decode(encoded_bytes)
    assert decoded == ""


def test_encode() -> None:
    subcon = MagicMock(spec=Construct)
    subcon.flagbuildnone = False
    string = "example"
    value = _Eq3Serial(subcon)
    encoded = value._encode(string, MagicMock(), "")
    assert encoded == bytes([ord(char) + 0x30 for char in string])


def test_decode() -> None:
    subcon = MagicMock(spec=Construct)
    subcon.flagbuildnone = False
    encoded = bytes([ord(char) + 0x30 for char in "example"])
    value = _Eq3Serial(subcon)
    decoded = value._decode(encoded, MagicMock(), "")
    assert decoded == "example"



from datetime import datetime, time, timedelta

from construct_typed import Adapter, Context

from eq3btsmart.const import (
    EQ3_MAX_OFFSET,
    EQ3_MIN_OFFSET,
    EQ3_OFF_TEMP,
    EQ3_ON_TEMP,
)
from eq3btsmart.exceptions import Eq3InvalidDataException

__all__: list[str] = []


class _Eq3AwayTime(Adapter[bytes, bytes, datetime, datetime]):
    """Adapter to encode and decode away time data."""

    @staticmethod
    def from_datetime(value: datetime) -> bytes:
        value += timedelta(minutes=15)
        value -= timedelta(minutes=value.minute % 30)

        if value.year < 2000 or value.year > 2099:
            raise Eq3InvalidDataException("Invalid year, possible [2000, 2099]")

        year = value.year - 2000
        hour = value.hour * 2
        if value.minute != 0:
            hour |= 0x01

        return bytes([value.day, year, hour, value.month])

    @staticmethod
    def to_datetime(value: bytes) -> datetime:
        if value == bytes([0x00, 0x00, 0x00, 0x00]):
            return datetime(year=2000, month=1, day=1, hour=0, minute=0)

        (day, year, hour_min, month) = value

        year += 2000

        minute = 0
        if hour_min & 0x01:
            minute = 30
        hour = int(hour_min / 2)

        return datetime(year=year, month=month, day=day, hour=hour, minute=minute)

    def _encode(self, obj: datetime, _ctx: Context, _path: str) -> bytes:
        return self.from_datetime(obj)

    def _decode(self, obj: bytes, _ctx: Context, _path: str) -> datetime:
        return self.to_datetime(obj)


class _Eq3Duration(Adapter[int, int, timedelta, timedelta]):
    """Adapter to encode and decode duration data."""

    def _encode(self, obj: timedelta, _ctx: Context, _path: str) -> int:
        return self.encode(obj)

    def _decode(self, obj: int, _ctx: Context, _path: str) -> timedelta:
        return self.decode(obj)

    @classmethod
    def encode(cls, value: timedelta) -> int:
        if value.seconds < 0 or value.seconds > 3600.0:
            raise Eq3InvalidDataException(
                "Window open time must be between 0 and 60 minutes in intervals of 5 minutes."
            )

        return int(value.seconds / 300.0)

    @classmethod
    def decode(cls, value: int) -> timedelta:
        return timedelta(minutes=float(value * 5.0))


class _Eq3ScheduleTime(Adapter[int, int, time, time]):
    """Adapter to encode and decode schedule time data."""

    def _encode(self, obj: time, _ctx: Context, _path: str) -> int:
        return self.encode(obj)

    def _decode(self, obj: int, _ctx: Context, _path: str) -> time:
        return self.decode(obj)

    @classmethod
    def encode(cls, value: time) -> int:
        return int((value.hour * 60 + value.minute) / 10)

    @classmethod
    def decode(cls, value: int) -> time:
        hour, minute = divmod(value * 10, 60)
        return time(hour=hour, minute=minute)


class _Eq3Serial(Adapter[bytes, bytes, str, str]):
    """Adapter to encode and decode serial id."""

    def _encode(self, obj: str, _ctx: Context, _path: str) -> bytes:
        return self.encode(obj)

    def _decode(self, obj: bytes, _ctx: Context, _path: str) -> str:
        return self.decode(obj)

    @classmethod
    def encode(cls, value: str) -> bytes:
        return bytes(char + 0x30 for char in value.encode())

    @classmethod
    def decode(cls, value: bytes) -> str:
        return bytes(char - 0x30 for char in value).decode()


class _Eq3TemperatureOffset(Adapter[int, int, float, float]):
    """Adapter to encode and decode temperature offset data."""

    def _encode(self, obj: float, _ctx: Context, _path: str) -> int:
        return self.encode(obj)

    def _decode(self, obj: int, _ctx: Context, _path: str) -> float:
        return self.decode(obj)

    @classmethod
    def encode(cls, value: float) -> int:
        if value < EQ3_MIN_OFFSET or value > EQ3_MAX_OFFSET:
            raise Eq3InvalidDataException(
                f"Temperature {value} out of range [{EQ3_MIN_OFFSET}, {EQ3_MAX_OFFSET}]"
            )

        return int((value + 3.5) / 0.5)

    @classmethod
    def decode(cls, value: int) -> float:
        return float(value * 0.5 - 3.5)


class _Eq3Temperature(Adapter[int, int, float, float]):
    """Adapter to encode and decode temperature data."""

    def _encode(self, obj: float, _ctx: Context, _path: str) -> int:
        return self.encode(obj)

    def _decode(self, obj: int, _ctx: Context, _path: str) -> float:
        return self.decode(obj)

    @classmethod
    def encode(cls, value: float) -> int:
        if value < EQ3_OFF_TEMP or value > EQ3_ON_TEMP:
            raise Eq3InvalidDataException(
                f"Temperature {value} out of range [{EQ3_OFF_TEMP}, {EQ3_ON_TEMP}]"
            )

        return int(value * 2)

    @classmethod
    def decode(cls, value: int) -> float:
        return float(value / 2)


class _Eq3Time(Adapter[bytes, bytes, datetime, datetime]):
    """Adapter to encode and decode time data."""

    def _encode(self, obj: datetime, _ctx: Context, _path: str) -> bytes:
        return self.encode(obj)

    def _decode(self, obj: bytes, _ctx: Context, _path: str) -> datetime:
        return self.decode(obj)

    @classmethod
    def encode(cls, value: datetime) -> bytes:
        return bytes(
            [
                value.year % 100,
                value.month,
                value.day,
                value.hour,
                value.minute,
                value.second,
            ]
        )

    @classmethod
    def decode(cls, value: bytes) -> datetime:
        try:
            (year, month, day, hour, minute, second) = value
        except ValueError:
            raise Eq3InvalidDataException("Invalid time data")

        year += 2000

        return datetime(
            year=year, month=month, day=day, hour=hour, minute=minute, second=second
        )



import asyncio
from datetime import datetime, time, timedelta
from unittest.mock import (
    AsyncMock,
    MagicMock,
    PropertyMock,
    call,
    create_autospec,
    patch,
)

import pytest
from bleak.exc import BleakError

from eq3btsmart._adapters import _Eq3Temperature
from eq3btsmart._structures import (
    _AwaySetCommand,
    _BoostSetCommand,
    _ComfortEcoConfigureCommand,
    _ComfortSetCommand,
    _DeviceDataStruct,
    _EcoSetCommand,
    _Eq3Message,
    _Eq3Struct,
    _IdGetCommand,
    _InfoGetCommand,
    _LockSetCommand,
    _ModeSetCommand,
    _OffsetConfigureCommand,
    _ScheduleDayStruct,
    _ScheduleHourStruct,
    _ScheduleSetCommand,
    _StatusStruct,
    _TemperatureSetCommand,
    _WindowOpenConfigureCommand,
)
from eq3btsmart.const import (
    EQ3_OFF_TEMP,
    EQ3_ON_TEMP,
    Eq3Event,
    Eq3OperationMode,
    Eq3Preset,
    Eq3WeekDay,
    _Eq3Characteristic,
    _Eq3Command,
    _Eq3StatusFlags,
)
from eq3btsmart.exceptions import (
    Eq3AlreadyAwaitingResponseException,
    Eq3CommandException,
    Eq3ConnectionException,
    Eq3InternalException,
    Eq3InvalidDataException,
    Eq3StateException,
    Eq3TimeoutException,
)
from eq3btsmart.models import DeviceData, Schedule, ScheduleDay, ScheduleHour, Status
from eq3btsmart.thermostat import Thermostat


@pytest.mark.asyncio
async def test_connect(thermostat: Thermostat) -> None:
    with (
        patch.object(
            thermostat._conn, "connect", new_callable=AsyncMock
        ) as mock_connect,
        patch.object(
            thermostat._conn, "start_notify", new_callable=AsyncMock
        ) as mock_start_notify,
        patch.object(
            thermostat, "async_get_status", new_callable=AsyncMock
        ) as mock_get_status,
        patch.object(
            thermostat, "async_get_device_data", new_callable=AsyncMock
        ) as mock_get_device_data,
        patch.object(
            thermostat, "async_get_schedule", new_callable=AsyncMock
        ) as mock_get_schedule,
        patch.object(
            thermostat, "_trigger_event", new_callable=AsyncMock
        ) as mock_trigger_event,
    ):
        await thermostat.async_connect()

        mock_connect.assert_called_once()
        mock_start_notify.assert_called_once()
        mock_get_status.assert_called_once()
        mock_get_device_data.assert_called_once()
        mock_get_schedule.assert_called_once()
        mock_trigger_event.assert_called_once_with(
            Eq3Event.CONNECTED,
            device_data=thermostat.device_data,
            status=thermostat.status,
            schedule=thermostat.schedule,
        )


@pytest.mark.asyncio
async def test_connect_already_connected(thermostat: Thermostat) -> None:
    with patch.object(
        Thermostat, "is_connected", new_callable=PropertyMock
    ) as mock_is_connected:
        mock_is_connected.return_value = True

        with pytest.raises(Eq3StateException, match="Already connected"):
            await thermostat.async_connect()


@pytest.mark.asyncio
async def test_connect_bleak_error(thermostat: Thermostat) -> None:
    with patch.object(thermostat._conn, "connect", side_effect=BleakError):
        with pytest.raises(
            Eq3ConnectionException, match="Could not connect to the device"
        ):
            await thermostat.async_connect()


@pytest.mark.asyncio
async def test_connect_timeout_error(thermostat: Thermostat) -> None:
    with patch.object(thermostat._conn, "connect", side_effect=TimeoutError):
        with pytest.raises(Eq3TimeoutException, match="Timeout during connection"):
            await thermostat.async_connect()


@pytest.mark.asyncio
async def test_disconnect(thermostat: Thermostat) -> None:
    with (
        patch.object(
            thermostat._conn, "disconnect", new_callable=AsyncMock
        ) as mock_disconnect,
        patch.object(
            Thermostat, "is_connected", new_callable=PropertyMock
        ) as mock_is_connected,
    ):
        mock_is_connected.return_value = True

        await thermostat.async_disconnect()

        mock_disconnect.assert_called_once()


@pytest.mark.asyncio
async def test_disconnect_not_connected(thermostat: Thermostat) -> None:
    with patch.object(
        Thermostat, "is_connected", new_callable=PropertyMock
    ) as mock_is_connected:
        mock_is_connected.return_value = False

        with pytest.raises(Eq3StateException, match="Not connected"):
            await thermostat.async_disconnect()


@pytest.mark.asyncio
async def test_disconnect_bleak_error(thermostat: Thermostat) -> None:
    with (
        patch.object(thermostat._conn, "disconnect", side_effect=BleakError),
        patch.object(
            Thermostat, "is_connected", new_callable=PropertyMock
        ) as mock_is_connected,
    ):
        mock_is_connected.return_value = True

        with pytest.raises(
            Eq3ConnectionException, match="Could not disconnect from the device"
        ):
            await thermostat.async_disconnect()


@pytest.mark.asyncio
async def test_disconnect_timeout_error(thermostat: Thermostat) -> None:
    with (
        patch.object(thermostat._conn, "disconnect", side_effect=TimeoutError),
        patch.object(
            Thermostat, "is_connected", new_callable=PropertyMock
        ) as mock_is_connected,
    ):
        mock_is_connected.return_value = True

        with pytest.raises(Eq3TimeoutException, match="Timeout during disconnection"):
            await thermostat.async_disconnect()


@pytest.mark.asyncio
async def test_disconnect_with_pending_futures(thermostat: Thermostat) -> None:
    thermostat._device_data_future = asyncio.Future()
    thermostat._status_future = asyncio.Future()
    thermostat._schedule_future = asyncio.Future()

    with (
        patch.object(thermostat._conn, "disconnect", new_callable=AsyncMock),
        patch.object(
            Thermostat, "is_connected", new_callable=PropertyMock
        ) as mock_is_connected,
    ):
        mock_is_connected.return_value = True

        await thermostat.async_disconnect()

        assert thermostat._device_data_future.done()
        assert thermostat._status_future.done()
        assert thermostat._schedule_future.done()
        assert isinstance(
            thermostat._device_data_future.exception(), Eq3ConnectionException
        )
        assert isinstance(thermostat._status_future.exception(), Eq3ConnectionException)
        assert isinstance(
            thermostat._schedule_future.exception(), Eq3ConnectionException
        )


@pytest.mark.asyncio
async def test_get_device_data(thermostat: Thermostat) -> None:
    with patch.object(
        thermostat,
        "_async_write_command_with_device_data_response",
        new_callable=AsyncMock,
    ) as mock_write_command:
        await thermostat.async_get_device_data()

        mock_write_command.assert_called_once_with(_IdGetCommand())


@pytest.mark.asyncio
async def test_get_status(thermostat: Thermostat) -> None:
    with (
        patch.object(
            thermostat,
            "_async_write_command_with_status_response",
            new_callable=AsyncMock,
        ) as mock_write_command,
        patch("eq3btsmart.thermostat.datetime", autospec=True) as mock_datetime,
    ):
        now = datetime.now()
        mock_datetime.now.return_value = now

        await thermostat.async_get_status()

        mock_write_command.assert_called_once_with(_InfoGetCommand(time=now))


@pytest.mark.asyncio
async def test_get_schedule(thermostat: Thermostat) -> None:
    with patch.object(
        thermostat,
        "_async_write_commands_with_schedule_response",
        new_callable=AsyncMock,
    ) as mock_write_command:
        await thermostat.async_get_schedule()

        assert mock_write_command.called
        assert len(mock_write_command.call_args.args[0]) == 7


@pytest.mark.asyncio
async def test_device_data(thermostat: Thermostat) -> None:
    device_data = MagicMock()
    thermostat._last_device_data = device_data

    assert thermostat.device_data == device_data


@pytest.mark.asyncio
async def test_device_data_not_set(thermostat: Thermostat) -> None:
    thermostat._last_device_data = None

    with pytest.raises(Eq3StateException, match="Device data not set"):
        _ = thermostat.device_data


@pytest.mark.asyncio
async def test_status(thermostat: Thermostat) -> None:
    status = MagicMock()
    thermostat._last_status = status

    assert thermostat.status == status


@pytest.mark.asyncio
async def test_status_not_set(thermostat: Thermostat) -> None:
    thermostat._last_status = None

    with pytest.raises(Eq3StateException, match="Status not set"):
        _ = thermostat.status


@pytest.mark.asyncio
async def test_presets(thermostat: Thermostat) -> None:
    thermostat._last_status = MagicMock()
    presets = MagicMock()
    thermostat._last_status.presets = presets

    assert thermostat.presets == presets


@pytest.mark.asyncio
async def test_presets_not_set(thermostat: Thermostat) -> None:
    thermostat._last_status = MagicMock()
    thermostat._last_status.presets = None

    with pytest.raises(Eq3StateException, match="Presets not set"):
        _ = thermostat.presets


@pytest.mark.asyncio
async def test_schedule(thermostat: Thermostat) -> None:
    schedule = MagicMock()
    thermostat._last_schedule = schedule

    assert thermostat.schedule == schedule


@pytest.mark.asyncio
async def test_schedule_not_set(thermostat: Thermostat) -> None:
    thermostat._last_schedule = None

    with pytest.raises(Eq3StateException, match="Schedule not set"):
        _ = thermostat.schedule


@pytest.mark.asyncio
async def test_set_temperature(thermostat: Thermostat) -> None:
    with patch.object(
        thermostat, "_async_write_command_with_status_response", new_callable=AsyncMock
    ) as mock_write_command:
        await thermostat.async_set_temperature(21.0)

        mock_write_command.assert_called_once_with(
            _TemperatureSetCommand(temperature=21.0)
        )


@pytest.mark.asyncio
async def test_set_temperature_off(thermostat: Thermostat) -> None:
    with (
        patch.object(
            thermostat,
            "_async_write_command_with_status_response",
            new_callable=AsyncMock,
        ) as mock_write_command,
        patch.object(
            thermostat, "async_set_mode", new_callable=AsyncMock
        ) as mock_set_mode,
    ):
        await thermostat.async_set_temperature(EQ3_OFF_TEMP)

        mock_write_command.assert_not_called()
        mock_set_mode.assert_called_once_with(Eq3OperationMode.OFF)


@pytest.mark.asyncio
async def test_set_temperature_on(thermostat: Thermostat) -> None:
    with (
        patch.object(
            thermostat,
            "_async_write_command_with_status_response",
            new_callable=AsyncMock,
        ) as mock_write_command,
        patch.object(
            thermostat, "async_set_mode", new_callable=AsyncMock
        ) as mock_set_mode,
    ):
        await thermostat.async_set_temperature(EQ3_ON_TEMP)

        mock_write_command.assert_not_called()
        mock_set_mode.assert_called_once_with(Eq3OperationMode.ON)


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "mode, command",
    [
        (
            Eq3OperationMode.MANUAL,
            _ModeSetCommand(mode=Eq3OperationMode.MANUAL | _Eq3Temperature.encode(20)),
        ),
        (
            Eq3OperationMode.AUTO,
            _ModeSetCommand(mode=Eq3OperationMode.AUTO),
        ),
        (
            Eq3OperationMode.ON,
            _ModeSetCommand(
                mode=Eq3OperationMode.MANUAL | _Eq3Temperature.encode(EQ3_ON_TEMP)
            ),
        ),
        (
            Eq3OperationMode.OFF,
            _ModeSetCommand(
                mode=Eq3OperationMode.MANUAL | _Eq3Temperature.encode(EQ3_OFF_TEMP)
            ),
        ),
    ],
)
async def test_set_mode(
    thermostat: Thermostat, mode: Eq3OperationMode, command: _ModeSetCommand
) -> None:
    with (
        patch.object(
            thermostat,
            "_async_write_command_with_status_response",
            new_callable=AsyncMock,
        ) as mock_write_command,
        patch.object(
            Thermostat, "status", new_callable=PropertyMock
        ) as mock_status_property,
    ):
        mock_status = MagicMock()
        mock_status.target_temperature = 20.0
        mock_status_property.return_value = mock_status
        await thermostat.async_set_mode(mode)

        mock_write_command.assert_called_once_with(command)


@pytest.mark.asyncio
async def test_set_invalid_mode(thermostat: Thermostat) -> None:
    with patch.object(thermostat, "_async_write_command_with_status_response"):
        with pytest.raises(Eq3InvalidDataException, match="Unsupported operation mode"):
            await thermostat.async_set_mode(Eq3OperationMode.AWAY)


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "preset, command",
    [
        (
            Eq3Preset.COMFORT,
            _ComfortSetCommand(),
        ),
        (
            Eq3Preset.ECO,
            _EcoSetCommand(),
        ),
    ],
)
async def test_set_preset(
    thermostat: Thermostat, preset: Eq3Preset, command: _Eq3Command
) -> None:
    with patch.object(
        thermostat, "_async_write_command_with_status_response", new_callable=AsyncMock
    ) as mock_write_command:
        await thermostat.async_set_preset(preset)

        mock_write_command.assert_called_once_with(command)


@pytest.mark.asyncio
@pytest.mark.parametrize("enable", [True, False])
async def test_set_boost(thermostat: Thermostat, enable: bool) -> None:
    with patch.object(
        thermostat, "_async_write_command_with_status_response", new_callable=AsyncMock
    ) as mock_write_command:
        await thermostat.async_set_boost(enable)

        mock_write_command.assert_called_once_with(_BoostSetCommand(enable=enable))


@pytest.mark.asyncio
@pytest.mark.parametrize("enable", [True, False])
async def test_set_locked(thermostat: Thermostat, enable: bool) -> None:
    with patch.object(
        thermostat, "_async_write_command_with_status_response", new_callable=AsyncMock
    ) as mock_write_command:
        await thermostat.async_set_locked(enable)

        mock_write_command.assert_called_once_with(_LockSetCommand(enable=enable))


@pytest.mark.asyncio
async def test_set_away(thermostat: Thermostat) -> None:
    with patch.object(
        thermostat, "_async_write_command_with_status_response", new_callable=AsyncMock
    ) as mock_write_command:
        now = datetime.now()
        await thermostat.async_set_away(now, 21.0)

        mock_write_command.assert_called_once_with(
            _AwaySetCommand(
                mode=Eq3OperationMode.AWAY | _Eq3Temperature.encode(21.0),
                away_until=now,
            )
        )


@pytest.mark.asyncio
async def test_set_schedule(thermostat: Thermostat) -> None:
    with patch.object(
        thermostat,
        "_async_write_commands_with_schedule_response",
        new_callable=AsyncMock,
    ) as mock_write_command:
        schedule = Schedule(
            schedule_days=[
                ScheduleDay(
                    Eq3WeekDay.TUESDAY,
                    schedule_hours=[
                        ScheduleHour(15.5, time(hour=1, minute=0)),
                        ScheduleHour(28.5, time(hour=23, minute=30)),
                    ],
                ),
                ScheduleDay(
                    Eq3WeekDay.WEDNESDAY,
                    schedule_hours=[
                        ScheduleHour(12.5, time(hour=12, minute=30)),
                        ScheduleHour(23.0, time(hour=15, minute=00)),
                    ],
                ),
            ]
        )
        await thermostat.async_set_schedule(schedule)

        mock_write_command.assert_called_once_with(
            [
                _ScheduleSetCommand(
                    day=schedule.schedule_days[0].week_day,
                    hours=[
                        _ScheduleHourStruct(
                            target_temp=schedule.schedule_days[0]
                            .schedule_hours[0]
                            .target_temperature,
                            next_change_at=schedule.schedule_days[0]
                            .schedule_hours[0]
                            .next_change_at,
                        ),
                        _ScheduleHourStruct(
                            target_temp=schedule.schedule_days[0]
                            .schedule_hours[1]
                            .target_temperature,
                            next_change_at=schedule.schedule_days[0]
                            .schedule_hours[1]
                            .next_change_at,
                        ),
                    ],
                ),
                _ScheduleSetCommand(
                    day=schedule.schedule_days[1].week_day,
                    hours=[
                        _ScheduleHourStruct(
                            target_temp=schedule.schedule_days[1]
                            .schedule_hours[0]
                            .target_temperature,
                            next_change_at=schedule.schedule_days[1]
                            .schedule_hours[0]
                            .next_change_at,
                        ),
                        _ScheduleHourStruct(
                            target_temp=schedule.schedule_days[1]
                            .schedule_hours[1]
                            .target_temperature,
                            next_change_at=schedule.schedule_days[1]
                            .schedule_hours[1]
                            .next_change_at,
                        ),
                    ],
                ),
            ]
        )


@pytest.mark.asyncio
async def test_delete_schedule(thermostat: Thermostat) -> None:
    with patch.object(
        thermostat,
        "_async_write_commands_with_schedule_response",
        new_callable=AsyncMock,
    ) as mock_write_command:
        await thermostat.async_delete_schedule()

        mock_write_command.assert_called_once_with(
            [_ScheduleSetCommand(day=day, hours=[]) for day in Eq3WeekDay]
        )


@pytest.mark.asyncio
async def test_delete_schedule_single_day(thermostat: Thermostat) -> None:
    with patch.object(
        thermostat,
        "_async_write_commands_with_schedule_response",
        new_callable=AsyncMock,
    ) as mock_write_command:
        await thermostat.async_delete_schedule(Eq3WeekDay.MONDAY)

        mock_write_command.assert_called_once_with(
            [_ScheduleSetCommand(day=Eq3WeekDay.MONDAY, hours=[])]
        )


@pytest.mark.asyncio
async def test_delete_schedule_multiple_days(thermostat: Thermostat) -> None:
    with patch.object(
        thermostat,
        "_async_write_commands_with_schedule_response",
        new_callable=AsyncMock,
    ) as mock_write_command:
        await thermostat.async_delete_schedule([Eq3WeekDay.MONDAY, Eq3WeekDay.TUESDAY])

        mock_write_command.assert_called_once_with(
            [
                _ScheduleSetCommand(day=Eq3WeekDay.MONDAY, hours=[]),
                _ScheduleSetCommand(day=Eq3WeekDay.TUESDAY, hours=[]),
            ]
        )


@pytest.mark.asyncio
async def test_configure_presets(thermostat: Thermostat) -> None:
    with patch.object(
        thermostat, "_async_write_command_with_status_response", new_callable=AsyncMock
    ) as mock_write_command:
        await thermostat.async_configure_presets(26.5, 16.0)

        mock_write_command.assert_called_once_with(
            _ComfortEcoConfigureCommand(
                comfort_temperature=26.5,
                eco_temperature=16.0,
            )
        )


@pytest.mark.asyncio
async def test_configure_comfort_temperature(thermostat: Thermostat) -> None:
    with (
        patch.object(
            thermostat,
            "_async_write_command_with_status_response",
            new_callable=AsyncMock,
        ) as mock_write_command,
        patch.object(
            Thermostat, "presets", new_callable=PropertyMock
        ) as mock_presets_property,
    ):
        mock_presets = MagicMock()
        mock_presets.eco_temperature = 19.5
        mock_presets_property.return_value = mock_presets

        await thermostat.async_configure_comfort_temperature(25.5)

        mock_write_command.assert_called_once_with(
            _ComfortEcoConfigureCommand(comfort_temperature=25.5, eco_temperature=19.5)
        )


@pytest.mark.asyncio
async def test_configure_eco_temperature(thermostat: Thermostat) -> None:
    with (
        patch.object(
            thermostat,
            "_async_write_command_with_status_response",
            new_callable=AsyncMock,
        ) as mock_write_command,
        patch.object(
            Thermostat, "presets", new_callable=PropertyMock
        ) as mock_presets_property,
    ):
        mock_presets = MagicMock()
        mock_presets.comfort_temperature = 19.5
        mock_presets_property.return_value = mock_presets

        await thermostat.async_configure_eco_temperature(15.5)

        mock_write_command.assert_called_once_with(
            _ComfortEcoConfigureCommand(comfort_temperature=19.5, eco_temperature=15.5)
        )


@pytest.mark.asyncio
async def test_configure_temperature_offset(thermostat: Thermostat) -> None:
    with patch.object(
        thermostat, "_async_write_command_with_status_response", new_callable=AsyncMock
    ) as mock_write_command:
        await thermostat.async_configure_temperature_offset(2.5)

        mock_write_command.assert_called_once_with(_OffsetConfigureCommand(offset=2.5))


@pytest.mark.asyncio
async def test_configure_window_open(thermostat: Thermostat) -> None:
    with patch.object(
        thermostat, "_async_write_command_with_status_response", new_callable=AsyncMock
    ) as mock_write_command:
        await thermostat.async_configure_window_open(6.5, timedelta(minutes=25))

        mock_write_command.assert_called_once_with(
            _WindowOpenConfigureCommand(
                window_open_temperature=6.5,
                window_open_time=timedelta(minutes=25),
            )
        )


@pytest.mark.asyncio
async def test_configure_window_open_float(thermostat: Thermostat) -> None:
    with patch.object(
        thermostat, "_async_write_command_with_status_response", new_callable=AsyncMock
    ) as mock_write_command:
        await thermostat.async_configure_window_open(6.5, 25)

        mock_write_command.assert_called_once_with(
            _WindowOpenConfigureCommand(
                window_open_temperature=6.5,
                window_open_time=timedelta(minutes=25),
            )
        )


@pytest.mark.asyncio
async def test_configure_window_open_temperature(thermostat: Thermostat) -> None:
    with (
        patch.object(
            thermostat,
            "_async_write_command_with_status_response",
            new_callable=AsyncMock,
        ) as mock_write_command,
        patch.object(
            Thermostat, "presets", new_callable=PropertyMock
        ) as mock_presets_property,
    ):
        mock_presets = MagicMock()
        mock_presets.window_open_time = timedelta(minutes=15)
        mock_presets_property.return_value = mock_presets

        await thermostat.async_configure_window_open_temperature(7.5)

        mock_write_command.assert_called_once_with(
            _WindowOpenConfigureCommand(
                window_open_temperature=7.5,
                window_open_time=timedelta(minutes=15),
            )
        )


@pytest.mark.asyncio
async def test_configure_window_open_duration(thermostat: Thermostat) -> None:
    with (
        patch.object(
            thermostat,
            "_async_write_command_with_status_response",
            new_callable=AsyncMock,
        ) as mock_write_command,
        patch.object(
            Thermostat, "presets", new_callable=PropertyMock
        ) as mock_presets_property,
    ):
        mock_presets = MagicMock()
        mock_presets.window_open_temperature = 7.5
        mock_presets_property.return_value = mock_presets

        await thermostat.async_configure_window_open_duration(timedelta(minutes=25))

        mock_write_command.assert_called_once_with(
            _WindowOpenConfigureCommand(
                window_open_temperature=7.5,
                window_open_time=timedelta(minutes=25),
            )
        )


@pytest.mark.asyncio
async def test_configure_window_open_duration_float(thermostat: Thermostat) -> None:
    with (
        patch.object(
            thermostat,
            "_async_write_command_with_status_response",
            new_callable=AsyncMock,
        ) as mock_write_command,
        patch.object(
            Thermostat, "presets", new_callable=PropertyMock
        ) as mock_presets_property,
    ):
        mock_presets = MagicMock()
        mock_presets.window_open_temperature = 7.5
        mock_presets_property.return_value = mock_presets

        await thermostat.async_configure_window_open_duration(25.0)

        mock_write_command.assert_called_once_with(
            _WindowOpenConfigureCommand(
                window_open_temperature=7.5,
                window_open_time=timedelta(minutes=25),
            )
        )


@pytest.mark.asyncio
async def test_aenter_aexit() -> None:
    with (
        patch.object(
            Thermostat, "async_connect", new_callable=AsyncMock
        ) as mock_connect,
        patch.object(
            Thermostat, "async_disconnect", new_callable=AsyncMock
        ) as mock_disconnect,
    ):
        async with Thermostat("00:11:22:33:44:55"):
            mock_connect.assert_called_once()

        mock_disconnect.assert_called_once()


@pytest.mark.asyncio
async def test_write_command_with_device_data_response(thermostat: Thermostat) -> None:
    mock_command = MagicMock()
    mock_device_data = MagicMock()

    with patch.object(
        thermostat, "_async_write_command", new_callable=AsyncMock
    ) as mock_write_command:

        async def resolve_future() -> None:
            assert thermostat._device_data_future is not None
            thermostat._device_data_future.set_result(mock_device_data)

        asyncio.create_task(resolve_future())

        result = await thermostat._async_write_command_with_device_data_response(
            mock_command
        )

        mock_write_command.assert_called_once_with(mock_command)
        assert result == mock_device_data
        assert thermostat._device_data_future is None


@pytest.mark.asyncio
async def test_write_command_with_device_data_response_already_awaiting_response(
    thermostat: Thermostat,
) -> None:
    mock_command = MagicMock()
    thermostat._device_data_future = AsyncMock()

    with patch.object(
        thermostat, "_async_write_command", new_callable=AsyncMock
    ) as mock_write_command:
        with pytest.raises(
            Eq3AlreadyAwaitingResponseException,
            match="Already awaiting a device data command response",
        ):
            await thermostat._async_write_command_with_device_data_response(
                mock_command
            )

        mock_write_command.assert_not_called()


@pytest.mark.asyncio
async def test_write_command_with_device_data_response_timeout(
    thermostat: Thermostat,
) -> None:
    mock_command = MagicMock()

    with patch.object(
        thermostat, "_async_write_command", new_callable=AsyncMock
    ) as mock_write_command:
        thermostat._command_timeout = 0

        with pytest.raises(
            Eq3TimeoutException, match="Timeout during device data command"
        ):
            await thermostat._async_write_command_with_device_data_response(
                mock_command
            )

        mock_write_command.assert_called_once_with(mock_command)
        assert thermostat._device_data_future is None


@pytest.mark.asyncio
async def test_write_command_with_status_response(thermostat: Thermostat) -> None:
    mock_command = MagicMock()
    mock_status = MagicMock()

    with patch.object(
        thermostat, "_async_write_command", new_callable=AsyncMock
    ) as mock_write_command:

        async def resolve_future() -> None:
            assert thermostat._status_future is not None
            thermostat._status_future.set_result(mock_status)

        asyncio.create_task(resolve_future())

        result = await thermostat._async_write_command_with_status_response(
            mock_command
        )

        mock_write_command.assert_called_once_with(mock_command)
        assert result == mock_status
        assert thermostat._status_future is None


@pytest.mark.asyncio
async def test_write_command_with_status_response_already_awaiting_response(
    thermostat: Thermostat,
) -> None:
    mock_command = MagicMock()
    thermostat._status_future = AsyncMock()

    with patch.object(
        thermostat, "_async_write_command", new_callable=AsyncMock
    ) as mock_write_command:
        with pytest.raises(
            Eq3AlreadyAwaitingResponseException,
            match="Already awaiting a status command response",
        ):
            await thermostat._async_write_command_with_status_response(mock_command)

        mock_write_command.assert_not_called()


@pytest.mark.asyncio
async def test_write_command_with_status_response_timeout(
    thermostat: Thermostat,
) -> None:
    mock_command = MagicMock()

    with patch.object(
        thermostat, "_async_write_command", new_callable=AsyncMock
    ) as mock_write_command:
        thermostat._command_timeout = 0

        with pytest.raises(Eq3TimeoutException, match="Timeout during status command"):
            await thermostat._async_write_command_with_status_response(mock_command)

        mock_write_command.assert_called_once_with(mock_command)
        assert thermostat._status_future is None


@pytest.mark.asyncio
async def test_write_commands_with_schedule_response(thermostat: Thermostat) -> None:
    mock_commands = [create_autospec(_Eq3Struct) for _ in range(7)]
    mock_schedule = MagicMock()

    with patch.object(
        thermostat, "_async_write_command", new_callable=AsyncMock
    ) as mock_write_command:

        async def resolve_future() -> None:
            assert thermostat._schedule_future is not None
            thermostat._schedule_future.set_result(mock_schedule)

        asyncio.create_task(resolve_future())

        result = await thermostat._async_write_commands_with_schedule_response(
            mock_commands
        )

        mock_write_command.assert_has_calls(
            [call(command) for command in mock_commands]
        )
        assert result == mock_schedule
        assert thermostat._schedule_future is None


@pytest.mark.asyncio
async def test_write_commands_with_schedule_response_already_awaiting_response(
    thermostat: Thermostat,
) -> None:
    mock_commands = [
        create_autospec(
            _Eq3Struct,
        )
        for _ in range(7)
    ]
    thermostat._schedule_future = AsyncMock()

    with patch.object(
        thermostat, "_async_write_command", new_callable=AsyncMock
    ) as mock_write_command:
        with pytest.raises(
            Eq3AlreadyAwaitingResponseException,
            match="Already awaiting a schedule command response",
        ):
            await thermostat._async_write_commands_with_schedule_response(mock_commands)

        mock_write_command.assert_not_called()


@pytest.mark.asyncio
async def test_write_commands_with_schedule_response_timeout(
    thermostat: Thermostat,
) -> None:
    mock_commands = [create_autospec(_Eq3Struct) for _ in range(7)]

    with patch.object(
        thermostat, "_async_write_command", new_callable=AsyncMock
    ) as mock_write_command:
        thermostat._command_timeout = 0

        with pytest.raises(
            Eq3TimeoutException, match="Timeout during schedule command"
        ):
            await thermostat._async_write_commands_with_schedule_response(mock_commands)

        mock_write_command.assert_has_calls(
            [call(command) for command in mock_commands]
        )
        assert thermostat._schedule_future is None


@pytest.mark.asyncio
async def test_write_command(thermostat: Thermostat) -> None:
    mock_command = MagicMock()

    with (
        patch.object(
            thermostat._conn, "write_gatt_char", new_callable=AsyncMock
        ) as mock_write_gatt_char,
        patch.object(
            Thermostat, "is_connected", new_callable=PropertyMock
        ) as mock_is_connected,
    ):
        mock_is_connected.return_value = True

        await thermostat._async_write_command(mock_command)

        mock_write_gatt_char.assert_called_once_with(
            _Eq3Characteristic.WRITE, mock_command.to_bytes()
        )


@pytest.mark.asyncio
async def test_write_command_not_connected(thermostat: Thermostat) -> None:
    mock_command = MagicMock()

    with patch.object(
        thermostat._conn, "write_gatt_char", new_callable=AsyncMock
    ) as mock_write_gatt_char:
        with pytest.raises(Eq3StateException, match="Not connected"):
            await thermostat._async_write_command(mock_command)

        mock_write_gatt_char.assert_not_called()


@pytest.mark.asyncio
async def test_write_command_bleak_error(thermostat: Thermostat) -> None:
    mock_command = MagicMock()

    with (
        patch.object(thermostat._conn, "write_gatt_char", side_effect=BleakError),
        patch.object(
            Thermostat, "is_connected", new_callable=PropertyMock
        ) as mock_is_connected,
    ):
        mock_is_connected.return_value = True

        with pytest.raises(Eq3CommandException, match="Error during write"):
            await thermostat._async_write_command(mock_command)


@pytest.mark.asyncio
async def test_write_command_timeout_error(thermostat: Thermostat) -> None:
    mock_command = MagicMock()

    with (
        patch.object(thermostat._conn, "write_gatt_char", side_effect=TimeoutError),
        patch.object(
            Thermostat, "is_connected", new_callable=PropertyMock
        ) as mock_is_connected,
    ):
        thermostat._command_timeout = 0
        mock_is_connected.return_value = True

        with pytest.raises(Eq3TimeoutException, match="Timeout during write"):
            await thermostat._async_write_command(mock_command)


@pytest.mark.asyncio
async def test_on_disconnected(thermostat: Thermostat) -> None:
    with patch.object(
        thermostat, "_trigger_event", new_callable=AsyncMock
    ) as mock_trigger_event:
        thermostat._on_disconnected(MagicMock())

        mock_trigger_event.assert_called_once_with(Eq3Event.DISCONNECTED)


@pytest.mark.asyncio
async def test_on_message_received_device_data(thermostat: Thermostat) -> None:
    command = _DeviceDataStruct(
        version=0, unknown_1=0, unknown_2=0, serial="0000000000", unknown_3=0
    )
    data = command.to_bytes()
    characteristic = MagicMock()

    with patch.object(
        thermostat, "_on_device_data_received", new_callable=AsyncMock
    ) as mock_on_device_data_received:
        await thermostat._on_message_received(characteristic, bytearray(data))

        mock_on_device_data_received.assert_called_once_with(
            DeviceData._from_bytes(data)
        )


@pytest.mark.asyncio
async def test_on_message_received_status(thermostat: Thermostat) -> None:
    command = _StatusStruct(mode=_Eq3StatusFlags.MANUAL, valve=0, target_temp=4.5)
    data = command.to_bytes()
    characteristic = MagicMock()

    with patch.object(
        thermostat, "_on_status_received", new_callable=AsyncMock
    ) as mock_on_status_received:
        await thermostat._on_message_received(characteristic, bytearray(data))

        mock_on_status_received.assert_called_once_with(Status._from_bytes(data))


@pytest.mark.asyncio
async def test_on_message_received_status_invalid_data(thermostat: Thermostat) -> None:
    command = _StatusStruct(
        mode=_Eq3StatusFlags.MANUAL,
        valve=0,
        target_temp=4.5,
    )
    data = bytearray(command.to_bytes())
    data[1] = 0
    characteristic = MagicMock()

    with patch.object(
        thermostat, "_on_status_received", new_callable=AsyncMock
    ) as mock_on_status_received:
        await thermostat._on_message_received(characteristic, data)

        mock_on_status_received.assert_not_called()


@pytest.mark.asyncio
async def test_on_message_received_schedule(thermostat: Thermostat) -> None:
    command = _ScheduleDayStruct(
        day=Eq3WeekDay.MONDAY,
        hours=[
            _ScheduleHourStruct(
                target_temp=20.5, next_change_at=time(hour=1, minute=0)
            ),
            _ScheduleHourStruct(
                target_temp=22.5, next_change_at=time(hour=23, minute=30)
            ),
        ],
    )
    data = command.to_bytes()
    characteristic = MagicMock()

    with patch.object(
        thermostat, "_on_schedule_received", new_callable=AsyncMock
    ) as mock_on_schedule_received:
        await thermostat._on_message_received(characteristic, bytearray(data))

        mock_on_schedule_received.assert_called_once_with(Schedule._from_bytes(data))


@pytest.mark.asyncio
async def test_on_message_received_unknown(thermostat: Thermostat) -> None:
    command = _Eq3Message(cmd=0xFF, is_status_command=False, data=b"")
    data = command.to_bytes()
    characteristic = MagicMock()

    with pytest.raises(Eq3InternalException, match="Unknown command"):
        await thermostat._on_message_received(characteristic, bytearray(data))


@pytest.mark.asyncio
async def test_on_device_data_received(thermostat: Thermostat) -> None:
    device_data = MagicMock()

    with patch.object(
        thermostat, "_trigger_event", new_callable=AsyncMock
    ) as mock_trigger_event:
        await thermostat._on_device_data_received(device_data)

        mock_trigger_event.assert_called_once_with(
            Eq3Event.DEVICE_DATA_RECEIVED, device_data=device_data
        )


@pytest.mark.asyncio
async def test_on_device_data_received_with_future(thermostat: Thermostat) -> None:
    device_data = MagicMock()
    thermostat._device_data_future = asyncio.Future()

    with patch.object(
        thermostat, "_trigger_event", new_callable=AsyncMock
    ) as mock_trigger_event:
        await thermostat._on_device_data_received(device_data)

        mock_trigger_event.assert_called_once_with(
            Eq3Event.DEVICE_DATA_RECEIVED, device_data=device_data
        )
        assert thermostat._device_data_future.done()
        assert thermostat._device_data_future.result() == device_data


@pytest.mark.asyncio
async def test_on_status_received(thermostat: Thermostat) -> None:
    status = MagicMock()

    with patch.object(
        thermostat, "_trigger_event", new_callable=AsyncMock
    ) as mock_trigger_event:
        await thermostat._on_status_received(status)

        mock_trigger_event.assert_called_once_with(
            Eq3Event.STATUS_RECEIVED, status=status
        )


@pytest.mark.asyncio
async def test_on_status_received_with_future(thermostat: Thermostat) -> None:
    status = MagicMock()
    thermostat._status_future = asyncio.Future()

    with patch.object(
        thermostat, "_trigger_event", new_callable=AsyncMock
    ) as mock_trigger_event:
        await thermostat._on_status_received(status)

        mock_trigger_event.assert_called_once_with(
            Eq3Event.STATUS_RECEIVED, status=status
        )
        assert thermostat._status_future.done()
        assert thermostat._status_future.result() == status


@pytest.mark.asyncio
async def test_on_schedule_received(thermostat: Thermostat) -> None:
    schedule = MagicMock()

    with patch.object(
        thermostat, "_trigger_event", new_callable=AsyncMock
    ) as mock_trigger_event:
        await thermostat._on_schedule_received(schedule)

        mock_trigger_event.assert_called_once_with(
            Eq3Event.SCHEDULE_RECEIVED, schedule=schedule
        )


@pytest.mark.asyncio
async def test_on_schedule_received_with_future(thermostat: Thermostat) -> None:
    schedule = MagicMock()
    thermostat._schedule_future = asyncio.Future()
    thermostat._schedule_future_counter = 1

    with patch.object(
        thermostat, "_trigger_event", new_callable=AsyncMock
    ) as mock_trigger_event:
        await thermostat._on_schedule_received(schedule)

        mock_trigger_event.assert_called_once_with(
            Eq3Event.SCHEDULE_RECEIVED, schedule=schedule
        )
        assert thermostat._schedule_future.done()
        assert thermostat._schedule_future.result() == schedule


@pytest.mark.asyncio
async def test_on_schedule_received_with_future_multiple(
    thermostat: Thermostat,
) -> None:
    schedule = MagicMock()
    thermostat._schedule_future = asyncio.Future()
    thermostat._schedule_future_counter = 2

    with patch.object(
        thermostat, "_trigger_event", new_callable=AsyncMock
    ) as mock_trigger_event:
        await thermostat._on_schedule_received(schedule)

        mock_trigger_event.assert_called_once_with(
            Eq3Event.SCHEDULE_RECEIVED, schedule=schedule
        )
        assert not thermostat._schedule_future.done()

        mock_trigger_event.reset_mock()

        await thermostat._on_schedule_received(schedule)

        mock_trigger_event.assert_called_once_with(
            Eq3Event.SCHEDULE_RECEIVED, schedule=schedule
        )
        assert thermostat._schedule_future.done()


@pytest.mark.asyncio
async def test_register_callback(thermostat: Thermostat) -> None:
    callback = lambda: None

    thermostat.register_callback(Eq3Event.DISCONNECTED, callback)

    assert thermostat._callbacks[Eq3Event.DISCONNECTED] == [callback]

    thermostat.register_callback(Eq3Event.DISCONNECTED, callback)

    assert thermostat._callbacks[Eq3Event.DISCONNECTED] == [callback]


@pytest.mark.asyncio
async def test_unregister_event(thermostat: Thermostat) -> None:
    callback = lambda: None

    assert thermostat._callbacks[Eq3Event.DISCONNECTED] == []

    thermostat.unregister_callback(Eq3Event.DISCONNECTED, callback)

    assert thermostat._callbacks[Eq3Event.DISCONNECTED] == []

    thermostat.register_callback(Eq3Event.DISCONNECTED, callback)

    assert thermostat._callbacks[Eq3Event.DISCONNECTED] == [callback]

    thermostat.unregister_callback(Eq3Event.DISCONNECTED, callback)

    assert thermostat._callbacks[Eq3Event.DISCONNECTED] == []


@pytest.mark.asyncio
async def test_trigger_event_connected(thermostat: Thermostat) -> None:
    device_data = DeviceData(321, "0000011111")
    status = Status(
        0, 20.5, Eq3OperationMode.AUTO, False, False, True, False, True, False
    )
    schedule = Schedule()
    callback = MagicMock()
    thermostat._callbacks[Eq3Event.CONNECTED] = [callback]

    await thermostat._trigger_event(
        Eq3Event.CONNECTED, device_data=device_data, status=status, schedule=schedule
    )

    callback.assert_called_once_with(device_data, status, schedule)


@pytest.mark.asyncio
async def test_trigger_event_connected_invalid_data(thermostat: Thermostat) -> None:
    device_data = DeviceData(321, "0000011111")
    status = Status(
        0, 20.5, Eq3OperationMode.AUTO, False, False, True, False, True, False
    )
    callback = MagicMock()
    thermostat._callbacks[Eq3Event.CONNECTED] = [callback]

    with pytest.raises(Eq3InternalException, match="must not be None for"):
        await thermostat._trigger_event(
            Eq3Event.CONNECTED,
            device_data=device_data,
            status=status,
            schedule=None,  # type: ignore
        )

    callback.assert_not_called()


@pytest.mark.asyncio
async def test_trigger_event_disconnected(thermostat: Thermostat) -> None:
    callback = MagicMock()
    thermostat._callbacks[Eq3Event.DISCONNECTED] = [callback]

    await thermostat._trigger_event(Eq3Event.DISCONNECTED)

    callback.assert_called_once()


@pytest.mark.asyncio
async def test_trigger_event_device_data_received(thermostat: Thermostat) -> None:
    device_data = DeviceData(321, "0000011111")
    callback = MagicMock()
    thermostat._callbacks[Eq3Event.DEVICE_DATA_RECEIVED] = [callback]

    await thermostat._trigger_event(
        Eq3Event.DEVICE_DATA_RECEIVED, device_data=device_data
    )

    callback.assert_called_once_with(device_data)


@pytest.mark.asyncio
async def test_trigger_event_device_data_received_invalid_data(
    thermostat: Thermostat,
) -> None:
    callback = MagicMock()
    thermostat._callbacks[Eq3Event.DEVICE_DATA_RECEIVED] = [callback]

    with pytest.raises(Eq3InternalException, match="must not be None for"):
        await thermostat._trigger_event(Eq3Event.DEVICE_DATA_RECEIVED, device_data=None)  # type: ignore

    callback.assert_not_called()


@pytest.mark.asyncio
async def test_trigger_event_status_received(thermostat: Thermostat) -> None:
    status = Status(
        0, 20.5, Eq3OperationMode.AUTO, False, False, True, False, True, False
    )
    callback = MagicMock()
    thermostat._callbacks[Eq3Event.STATUS_RECEIVED] = [callback]

    await thermostat._trigger_event(Eq3Event.STATUS_RECEIVED, status=status)

    callback.assert_called_once_with(status)


@pytest.mark.asyncio
async def test_trigger_event_status_received_invalid_data(
    thermostat: Thermostat,
) -> None:
    callback = MagicMock()
    thermostat._callbacks[Eq3Event.STATUS_RECEIVED] = [callback]

    with pytest.raises(Eq3InternalException, match="must not be None for"):
        await thermostat._trigger_event(Eq3Event.STATUS_RECEIVED, status=None)  # type: ignore

    callback.assert_not_called()


@pytest.mark.asyncio
async def test_trigger_event_schedule_received(thermostat: Thermostat) -> None:
    schedule = Schedule()
    callback = MagicMock()
    thermostat._callbacks[Eq3Event.SCHEDULE_RECEIVED] = [callback]

    await thermostat._trigger_event(Eq3Event.SCHEDULE_RECEIVED, schedule=schedule)

    callback.assert_called_once_with(schedule)


@pytest.mark.asyncio
async def test_trigger_event_schedule_received_invalid_data(
    thermostat: Thermostat,
) -> None:
    callback = MagicMock()
    thermostat._callbacks[Eq3Event.SCHEDULE_RECEIVED] = [callback]

    with pytest.raises(Eq3InternalException, match="must not be None for"):
        await thermostat._trigger_event(Eq3Event.SCHEDULE_RECEIVED, schedule=None)  # type: ignore

    callback.assert_not_called()


@pytest.mark.asyncio
async def tests_trigger_event_connected_async(thermostat: Thermostat) -> None:
    device_data = DeviceData(321, "0000011111")
    status = Status(
        0, 20.5, Eq3OperationMode.AUTO, False, False, True, False, True, False
    )
    schedule = Schedule()
    callback = AsyncMock()
    thermostat._callbacks[Eq3Event.CONNECTED] = [callback]

    await thermostat._trigger_event(
        Eq3Event.CONNECTED, device_data=device_data, status=status, schedule=schedule
    )

    callback.assert_called_once_with(device_data, status, schedule)



from datetime import datetime, timedelta
from unittest.mock import MagicMock

import pytest
from construct import Construct

from eq3btsmart._adapters import _Eq3AwayTime
from eq3btsmart.exceptions import Eq3InvalidDataException


def test_encode_valid_datetime() -> None:
    dt = datetime(year=2022, month=6, day=15, hour=10, minute=45)
    encoded = _Eq3AwayTime.from_datetime(dt)
    expected_bytes = bytes([15, 22, 22, 6])
    assert encoded == expected_bytes


def test_encode_valid_datetime_with_minute() -> None:
    dt = datetime(year=2022, month=6, day=15, hour=10, minute=25)
    encoded = _Eq3AwayTime.from_datetime(dt)
    expected_bytes = bytes([15, 22, 21, 6])
    assert encoded == expected_bytes


def test_decode_valid_bytes() -> None:
    encoded_bytes = bytes([15, 22, 21, 6])
    decoded = _Eq3AwayTime.to_datetime(encoded_bytes)
    expected_datetime = datetime(year=2022, month=6, day=15, hour=10, minute=30)
    assert decoded == expected_datetime


def test_round_trip_consistency() -> None:
    dt = datetime(year=2022, month=6, day=15, hour=10, minute=45)
    encoded = _Eq3AwayTime.from_datetime(dt)
    decoded = _Eq3AwayTime.to_datetime(encoded)
    dt_adjusted = dt + timedelta(minutes=15)
    dt_adjusted -= timedelta(minutes=dt_adjusted.minute % 30)
    assert decoded == dt_adjusted


def test_decode_special_case() -> None:
    encoded_bytes = bytes([0x00, 0x00, 0x00, 0x00])
    decoded = _Eq3AwayTime.to_datetime(encoded_bytes)
    expected_datetime = datetime(year=2000, month=1, day=1, hour=0, minute=0)
    assert decoded == expected_datetime


def test_encode_invalid_year() -> None:
    dt = datetime(year=2100, month=6, day=15, hour=10, minute=45)
    with pytest.raises(Eq3InvalidDataException):
        _Eq3AwayTime.from_datetime(dt)


def test_encode() -> None:
    subcon = MagicMock(spec=Construct)
    subcon.flagbuildnone = False
    dt = datetime(year=2022, month=6, day=15, hour=10, minute=45)
    value = _Eq3AwayTime(subcon)
    encoded = value._encode(dt, MagicMock(), "")
    expected_bytes = bytes([15, 22, 22, 6])
    assert encoded == expected_bytes


def test_decode() -> None:
    subcon = MagicMock(spec=Construct)
    subcon.flagbuildnone = False
    encoded_bytes = bytes([15, 22, 21, 6])
    value = _Eq3AwayTime(subcon)
    decoded = value._decode(encoded_bytes, MagicMock(), "")
    expected_datetime = datetime(year=2022, month=6, day=15, hour=10, minute=30)
    assert decoded == expected_datetime



from datetime import timedelta
from unittest.mock import MagicMock

import pytest
from construct import Construct

from eq3btsmart._adapters import _Eq3Duration
from eq3btsmart.exceptions import Eq3InvalidDataException


def test_encode_valid_timedelta() -> None:
    td = timedelta(minutes=25)
    encoded = _Eq3Duration.encode(td)
    assert encoded == 5


def test_decode_valid_int() -> None:
    encoded_int = 5
    decoded = _Eq3Duration.decode(encoded_int)
    assert decoded == timedelta(minutes=25)


def test_round_trip_consistency() -> None:
    td = timedelta(minutes=25)
    encoded = _Eq3Duration.encode(td)
    decoded = _Eq3Duration.decode(encoded)
    assert decoded == td


def test_encode_invalid_timedelta_negative() -> None:
    td = timedelta(minutes=-5)
    with pytest.raises(Eq3InvalidDataException):
        _Eq3Duration.encode(td)


def test_encode_invalid_timedelta_exceed() -> None:
    td = timedelta(minutes=65)
    with pytest.raises(Eq3InvalidDataException):
        _Eq3Duration.encode(td)


def test_encode() -> None:
    subcon = MagicMock(spec=Construct)
    subcon.flagbuildnone = False
    td = timedelta(minutes=25)
    value = _Eq3Duration(subcon)
    encoded = value._encode(td, MagicMock(), "")
    assert encoded == 5


def test_decode() -> None:
    subcon = MagicMock(spec=Construct)
    subcon.flagbuildnone = False
    encoded = 5
    value = _Eq3Duration(subcon)
    decoded = value._decode(encoded, MagicMock(), "")
    assert decoded == timedelta(minutes=25)



from datetime import datetime
from unittest.mock import MagicMock

import pytest
from construct import Construct

from eq3btsmart._adapters import _Eq3Time
from eq3btsmart.exceptions import Eq3InvalidDataException


def test_encode_valid_datetime() -> None:
    dt = datetime(year=2022, month=6, day=15, hour=10, minute=45, second=30)
    encoded = _Eq3Time.encode(dt)
    assert encoded == bytes([22, 6, 15, 10, 45, 30])


def test_decode_valid_bytes() -> None:
    encoded_bytes = bytes([22, 6, 15, 10, 45, 30])
    decoded = _Eq3Time.decode(encoded_bytes)
    assert decoded == datetime(
        year=2022, month=6, day=15, hour=10, minute=45, second=30
    )


def test_round_trip_consistency() -> None:
    dt = datetime(year=2022, month=6, day=15, hour=10, minute=45, second=30)
    encoded = _Eq3Time.encode(dt)
    decoded = _Eq3Time.decode(encoded)
    assert decoded == dt


def test_decode_invalid_bytes_length() -> None:
    encoded_bytes = bytes([22, 6, 15, 10, 45])
    with pytest.raises(Eq3InvalidDataException):
        _Eq3Time.decode(encoded_bytes)


def test_encode_future_date() -> None:
    dt = datetime(year=2122, month=6, day=15, hour=10, minute=45, second=30)
    encoded = _Eq3Time.encode(dt)
    assert encoded == bytes([22, 6, 15, 10, 45, 30])


def test_encode() -> None:
    subcon = MagicMock(spec=Construct)
    subcon.flagbuildnone = False
    dt = datetime(year=2022, month=6, day=15, hour=10, minute=45, second=30)
    value = _Eq3Time(subcon)
    encoded = value._encode(dt, MagicMock(), "")
    assert encoded == bytes([22, 6, 15, 10, 45, 30])


def test_decode() -> None:
    subcon = MagicMock(spec=Construct)
    subcon.flagbuildnone = False
    encoded = bytes([22, 6, 15, 10, 45, 30])
    value = _Eq3Time(subcon)
    decoded = value._decode(encoded, MagicMock(), "")
    assert decoded == datetime(
        year=2022, month=6, day=15, hour=10, minute=45, second=30
    )



from eq3btsmart.const import Eq3WeekDay


def test_from_index() -> None:
    assert Eq3WeekDay.from_index(0) == Eq3WeekDay.MONDAY
    assert Eq3WeekDay.from_index(1) == Eq3WeekDay.TUESDAY
    assert Eq3WeekDay.from_index(2) == Eq3WeekDay.WEDNESDAY
    assert Eq3WeekDay.from_index(3) == Eq3WeekDay.THURSDAY
    assert Eq3WeekDay.from_index(4) == Eq3WeekDay.FRIDAY
    assert Eq3WeekDay.from_index(5) == Eq3WeekDay.SATURDAY
    assert Eq3WeekDay.from_index(6) == Eq3WeekDay.SUNDAY



from datetime import time
from unittest.mock import MagicMock

from construct import Construct

from eq3btsmart._adapters import _Eq3ScheduleTime


def test_encode_valid_time() -> None:
    t = time(hour=13, minute=30)
    encoded = _Eq3ScheduleTime.encode(t)
    assert encoded == 81


def test_decode_valid_int() -> None:
    encoded_int = 81
    decoded = _Eq3ScheduleTime.decode(encoded_int)
    assert decoded == time(hour=13, minute=30)


def test_round_trip_consistency() -> None:
    t = time(hour=13, minute=30)
    encoded = _Eq3ScheduleTime.encode(t)
    decoded = _Eq3ScheduleTime.decode(encoded)
    assert decoded == t


def test_encode_edge_case() -> None:
    t = time(hour=23, minute=59)
    encoded = _Eq3ScheduleTime.encode(t)
    assert encoded == 143


def test_decode_edge_case() -> None:
    encoded_int = 143
    decoded = _Eq3ScheduleTime.decode(encoded_int)
    assert decoded == time(hour=23, minute=50)


def test_encode() -> None:
    subcon = MagicMock(spec=Construct)
    subcon.flagbuildnone = False
    t = time(hour=13, minute=30)
    value = _Eq3ScheduleTime(subcon)
    encoded = value._encode(t, MagicMock(), "")
    assert encoded == 81


def test_decode() -> None:
    subcon = MagicMock(spec=Construct)
    subcon.flagbuildnone = False
    encoded = 81
    value = _Eq3ScheduleTime(subcon)
    decoded = value._decode(encoded, MagicMock(), "")
    assert decoded == time(hour=13, minute=30)



"""Support for eQ-3 Bluetooth Smart thermostats."""

import asyncio
from collections import defaultdict
from datetime import datetime, timedelta
from types import TracebackType
from typing import Awaitable, Callable, Literal, Self, Union, overload

from bleak import BleakClient
from bleak.backends.characteristic import BleakGATTCharacteristic
from bleak.exc import BleakError
from construct_typed import DataclassStruct

from eq3btsmart._adapters import _Eq3Temperature
from eq3btsmart._structures import (
    _AwaySetCommand,
    _BoostSetCommand,
    _ComfortEcoConfigureCommand,
    _ComfortSetCommand,
    _EcoSetCommand,
    _Eq3Message,
    _Eq3Struct,
    _IdGetCommand,
    _InfoGetCommand,
    _LockSetCommand,
    _ModeSetCommand,
    _OffsetConfigureCommand,
    _ScheduleGetCommand,
    _ScheduleHourStruct,
    _ScheduleSetCommand,
    _TemperatureSetCommand,
    _WindowOpenConfigureCommand,
)
from eq3btsmart.const import (
    DEFAULT_COMMAND_TIMEOUT,
    DEFAULT_CONNECTION_TIMEOUT,
    EQ3_OFF_TEMP,
    EQ3_ON_TEMP,
    Eq3Event,
    Eq3OperationMode,
    Eq3Preset,
    Eq3WeekDay,
    _Eq3Characteristic,
    _Eq3Command,
)
from eq3btsmart.exceptions import (
    Eq3AlreadyAwaitingResponseException,
    Eq3CommandException,
    Eq3ConnectionException,
    Eq3InternalException,
    Eq3InvalidDataException,
    Eq3StateException,
    Eq3TimeoutException,
)
from eq3btsmart.models import DeviceData, Presets, Schedule, Status

__all__ = ["Thermostat"]


class Thermostat:
    """Representation of an eQ-3 Bluetooth Smart thermostat."""

    def __init__(
        self,
        mac_address: str,
        connection_timeout: int = DEFAULT_CONNECTION_TIMEOUT,
        command_timeout: int = DEFAULT_COMMAND_TIMEOUT,
    ):
        """Initialize the thermostat.

        The thermostat will be in a disconnected state after initialization.

        Args:
            mac_address (str): The MAC address of the thermostat.
            connection_timeout (int, optional): The connection timeout in seconds. Defaults to DEFAULT_CONNECTION_TIMEOUT.
            command_timeout (int, optional): The command timeout in seconds. Defaults to DEFAULT_COMMAND_TIMEOUT.
        """
        self._mac_address = mac_address

        self._last_status: Status | None = None
        self._last_device_data: DeviceData | None = None
        self._last_schedule: Schedule | None = None

        self._callbacks: defaultdict[
            Eq3Event, list[Union[Callable[..., None], Callable[..., Awaitable[None]]]]
        ] = defaultdict(list)
        self._conn: BleakClient = BleakClient(
            mac_address,
            disconnected_callback=self._on_disconnected,
            timeout=DEFAULT_CONNECTION_TIMEOUT,
        )
        self._device_data_future: asyncio.Future[DeviceData] | None = None
        self._status_future: asyncio.Future[Status] | None = None
        self._schedule_future: asyncio.Future[Schedule] | None = None
        self._schedule_future_counter: int = 0
        self._lock = asyncio.Lock()
        self._connection_timeout = connection_timeout
        self._command_timeout = command_timeout

    @property
    def is_connected(self) -> bool:
        """Check if the thermostat is connected.

        Returns:
            bool: True if connected, False otherwise.
        """
        return self._conn.is_connected

    @property
    def device_data(self) -> DeviceData:
        """Get the last known device data, ensuring it's not None.

        Returns:
            DeviceData: The last known device data.

        Raises:
            Eq3StateException: If the device data is None. This occurs when the thermostat has not been connected yet.
        """
        if self._last_device_data is None:
            raise Eq3StateException("Device data not set")
        return self._last_device_data

    @property
    def status(self) -> Status:
        """Get the last known status, ensuring it's not None.

        Returns:
            Status: The last known status.

        Raises:
            Eq3StateException: If the status is None. This occurs when the thermostat has not been connected yet.
        """
        if self._last_status is None:
            raise Eq3StateException("Status not set")
        return self._last_status

    @property
    def presets(self) -> Presets:
        """Get the presets from the last known status.

        Presets are only available since firmware version MIN_PRESETS_FW_VERSION.

        Returns:
            Presets: The presets.

        Raises:
            Eq3StateException: If the presets are None. This occurs when the thermostat has not been connected yet or if the thermostat's firmware does not support presets.
        """
        if self.status.presets is None:
            raise Eq3StateException("Presets not set")
        return self.status.presets

    @property
    def schedule(self) -> Schedule:
        """Get the last known schedule, ensuring it's not None.

        Returns:
            Schedule: The last known schedule.

        Raises:
            Eq3StateException: If the schedule is None. This occurs when the thermostat has not been connected
        """
        if self._last_schedule is None:
            raise Eq3StateException("Schedule not set")
        return self._last_schedule

    async def async_connect(self) -> None:
        """Connect to the thermostat.

        After connecting, the device data, status, and schedule will be queried and stored.
        When the connection is established, the CONNECTED event will be triggered.

        Raises:
            Eq3StateException: If the thermostat is already connected.
            Eq3ConnectionException: If the connection fails.
            Eq3TimeoutException: If the connection times out.
            Eq3CommandException: If an error occurs while sending a command.
        """
        if self.is_connected:
            raise Eq3StateException("Already connected")

        try:
            await asyncio.wait_for(self._conn.connect(), self._connection_timeout)
            await self._conn.start_notify(
                _Eq3Characteristic.NOTIFY, self._on_message_received
            )
            (
                self._last_device_data,
                self._last_status,
                self._last_schedule,
            ) = await asyncio.gather(
                self.async_get_device_data(),
                self.async_get_status(),
                self.async_get_schedule(),
            )
        except BleakError as ex:
            raise Eq3ConnectionException("Could not connect to the device") from ex
        except TimeoutError as ex:
            raise Eq3TimeoutException("Timeout during connection") from ex

        await self._trigger_event(
            Eq3Event.CONNECTED,
            device_data=self.device_data,
            status=self.status,
            schedule=self.schedule,
        )

    async def async_disconnect(self) -> None:
        """Disconnect from the thermostat.

        Before disconnection all pending futures will be cancelled.
        When the disconnection is complete, the DISCONNECTED event will be triggered.

        Raises:
            Eq3StateException: If the thermostat is not connected.
            Eq3ConnectionException: If the disconnection fails.
            Eq3TimeoutException: If the disconnection times out.
        """
        if not self.is_connected:
            raise Eq3StateException("Not connected")

        exception = Eq3ConnectionException("Connection closed")
        if self._device_data_future is not None and not self._device_data_future.done():
            self._device_data_future.set_exception(exception)

        if self._status_future is not None and not self._status_future.done():
            self._status_future.set_exception(exception)

        if self._schedule_future is not None and not self._schedule_future.done():
            self._schedule_future.set_exception(exception)

        try:
            await self._conn.disconnect()
        except BleakError as ex:
            raise Eq3ConnectionException("Could not disconnect from the device") from ex
        except TimeoutError as ex:
            raise Eq3TimeoutException("Timeout during disconnection") from ex

    async def async_get_device_data(self) -> DeviceData:
        """Query the latest device data.

        Returns:
            DeviceData: The device data.

        Raises:
            Eq3StateException: If the thermostat is not connected.
            Eq3CommandException: If an error occurs while sending the command.
            Eq3TimeoutException: If the command times out.
            Eq3AlreadyAwaitingResponseException: If a device data command is already pending.
        """
        return await self._async_write_command_with_device_data_response(
            _IdGetCommand()
        )

    async def async_get_status(self) -> Status:
        """Query the latest status.

        Returns:
            Status: The status.

        Raises:
            Eq3StateException: If the thermostat is not connected.
            Eq3CommandException: If an error occurs while sending the command.
            Eq3TimeoutException: If the command times out.
            Eq3AlreadyAwaitingResponseException: If a status command is already pending.
        """
        return await self._async_write_command_with_status_response(
            _InfoGetCommand(time=datetime.now())
        )

    async def async_get_schedule(self) -> Schedule:
        """Query the schedule.

        Returns:
            Schedule: The schedule.

        Raises:
            Eq3StateException: If the thermostat is not connected.
            Eq3CommandException: If an error occurs while sending the command.
            Eq3TimeoutException: If the command times out.
            Eq3AlreadyAwaitingResponseException: If a schedule command is already pending.
        """
        return await self._async_write_commands_with_schedule_response(
            [_ScheduleGetCommand(day=week_day) for week_day in Eq3WeekDay]
        )

    async def async_set_temperature(self, temperature: float) -> Status:
        """Set the target temperature.

        Temperatures are in degrees Celsius and specified in 0.5 degree increments.
        If the temperature is EQ3_OFF_TEMP, the thermostat will be turned off.
        If the temperature is EQ3_ON_TEMP, the thermostat will be turned on.

        Args:
            temperature (float): The new target temperature in degrees Celsius.

        Returns:
            Status: The updated status.

        Raises:
            Eq3StateException: If the thermostat is not connected.
            Eq3AlreadyAwaitingResponseException: If a status command is already pending.
            Eq3CommandException: If an error occurs during the command.
            Eq3TimeoutException: If the command times out.
            Eq3InvalidDataException: If the temperature is invalid.
        """
        if temperature == EQ3_OFF_TEMP:
            return await self.async_set_mode(Eq3OperationMode.OFF)

        if temperature == EQ3_ON_TEMP:
            return await self.async_set_mode(Eq3OperationMode.ON)

        return await self._async_write_command_with_status_response(
            _TemperatureSetCommand(temperature=temperature)
        )

    async def async_set_mode(self, operation_mode: Eq3OperationMode) -> Status:
        """Set the operation mode.

        Args:
            operation_mode (Eq3OperationMode): The new operation mode.

        Returns:
            Status: The updated status.

        Raises:
            Eq3StateException: If the thermostat is not connected.
            Eq3AlreadyAwaitingResponseException: If a status command is already pending.
            Eq3CommandException: If an error occurs during the command.
            Eq3TimeoutException: If the command times out.
            Eq3InvalidDataException: If the operation mode is not supported.
        """
        command: _ModeSetCommand

        match operation_mode:
            case Eq3OperationMode.AUTO:
                command = _ModeSetCommand(mode=Eq3OperationMode.AUTO)
            case Eq3OperationMode.MANUAL:
                command = _ModeSetCommand(
                    mode=Eq3OperationMode.MANUAL
                    | _Eq3Temperature.encode(self.status.target_temperature)
                )
            case Eq3OperationMode.OFF:
                command = _ModeSetCommand(
                    mode=Eq3OperationMode.MANUAL | _Eq3Temperature.encode(EQ3_OFF_TEMP)
                )
            case Eq3OperationMode.ON:
                command = _ModeSetCommand(
                    mode=Eq3OperationMode.MANUAL | _Eq3Temperature.encode(EQ3_ON_TEMP)
                )
            case _:
                raise Eq3InvalidDataException(
                    f"Unsupported operation mode: {operation_mode}"
                )

        return await self._async_write_command_with_status_response(command)

    async def async_set_preset(self, preset: Eq3Preset) -> Status:
        """Activate a preset.

        Args:
            preset (Eq3Preset): The preset to activate.

        Returns:
            Status: The updated status.

        Raises:
            Eq3StateException: If the thermostat is not connected.
            Eq3AlreadyAwaitingResponseException: If a status command is already pending.
            Eq3CommandException: If an error occurs during the command.
            Eq3TimeoutException: If the command times out.
        """
        command: _ComfortSetCommand | _EcoSetCommand

        match preset:
            case Eq3Preset.COMFORT:
                command = _ComfortSetCommand()
            case Eq3Preset.ECO:
                command = _EcoSetCommand()

        return await self._async_write_command_with_status_response(command)

    async def async_set_boost(self, enable: bool) -> Status:
        """Enable or disable the boost mode.

        The boost mode will set the target temperature to EQ3_ON_TEMP for 300 seconds and then revert to the previous target temperature.

        Args:
            enable (bool): True to enable the boost mode, False to disable it.

        Returns:
            Status: The updated status.

        Raises:
            Eq3StateException: If the thermostat is not connected.
            Eq3AlreadyAwaitingResponseException: If a status command is already pending.
            Eq3CommandException: If an error occurs during the command.
            Eq3TimeoutException: If the command times out.
        """
        return await self._async_write_command_with_status_response(
            _BoostSetCommand(enable=enable)
        )

    async def async_set_locked(self, enable: bool) -> Status:
        """Lock or unlock the thermostat.

        When locked, the thermostat's manual controls are disabled.

        Args:
            enable (bool): True to lock the thermostat, False to unlock it.

        Returns:
            Status: The updated status.

        Raises:
            Eq3StateException: If the thermostat is not connected.
            Eq3AlreadyAwaitingResponseException: If a status command is already pending.
            Eq3CommandException: If an error occurs during the command.
            Eq3TimeoutException: If the command times out.
        """
        return await self._async_write_command_with_status_response(
            _LockSetCommand(enable=enable)
        )

    async def async_set_away(
        self,
        away_until: datetime,
        temperature: float,
    ) -> Status:
        """Set the thermostat to away mode.

        The thermostat will be set to away mode until the specified date and time.
        The target temperature will be set to the specified temperature.
        Temperatures are in degrees Celsius and specified in 0.5 degree increments.

        Args:
            away_until (datetime): The date and time until the thermostat should be in away mode.
            temperature (float): The target temperature in degrees Celsius.

        Returns:
            Status: The updated status.

        Raises:
            Eq3StateException: If the thermostat is not connected.
            Eq3AlreadyAwaitingResponseException: If a status command is already pending.
            Eq3CommandException: If an error occurs during the command.
            Eq3TimeoutException: If the command times out.
            Eq3InvalidDataException: If the temperature or date is invalid.
        """
        return await self._async_write_command_with_status_response(
            _AwaySetCommand(
                mode=Eq3OperationMode.AWAY | _Eq3Temperature.encode(temperature),
                away_until=away_until,
            )
        )

    async def async_set_schedule(self, schedule: Schedule) -> Schedule:
        """Set the schedule.

        The schedule allows setting different target temperatures for each day of the week and different times of the day.
        It is only applied when the thermostat is in AUTO mode.

        Args:
            schedule: The schedule to set.

        Returns:
            The updated schedule.

        Raises:
            Eq3StateException: If the thermostat is not connected.
            Eq3AlreadyAwaitingResponseException: If a schedule command is already pending.
            Eq3CommandException: If an error occurs during the command.
            Eq3TimeoutException: If the command times out.
            Eq3InvalidDataException: If any of the schedule data is invalid.
        """
        return await self._async_write_commands_with_schedule_response(
            [
                _ScheduleSetCommand(
                    day=schedule_day.week_day,
                    hours=[
                        _ScheduleHourStruct(
                            target_temp=schedule_hour.target_temperature,
                            next_change_at=schedule_hour.next_change_at,
                        )
                        for schedule_hour in schedule_day.schedule_hours
                    ],
                )
                for schedule_day in schedule.schedule_days
            ]
        )

    async def async_delete_schedule(
        self, week_days: list[Eq3WeekDay] | Eq3WeekDay | None = None
    ) -> Schedule:
        """Delete the schedule for the specified week days.

        If no week days are specified, the schedule for all week days will be deleted.

        Args:
            week_days (list[WeekDay] | WeekDay | None, optional): The week days for which the schedule should be deleted.
                Can be a single WeekDay, a list of WeekDay, or None. Defaults to None.

        Returns:
            Schedule: The updated schedule after deletion.

        Raises:
            Eq3StateException: If the thermostat is not connected.
            Eq3AlreadyAwaitingResponseException: If a schedule command is already pending.
            Eq3CommandException: If an error occurs during the command.
            Eq3TimeoutException: If the command times out.
        """
        week_days = (
            [week_days]
            if isinstance(week_days, Eq3WeekDay)
            else week_days or list(Eq3WeekDay)
        )

        return await self._async_write_commands_with_schedule_response(
            [_ScheduleSetCommand(day=week_day, hours=[]) for week_day in week_days]
        )

    async def async_configure_presets(
        self,
        comfort_temperature: float,
        eco_temperature: float,
    ) -> Status:
        """Set the thermostat's preset temperatures.

        Temperatures are in degrees Celsius and specified in 0.5 degree increments.
        The initial values are 21.0 degrees Celsius for comfort and 17.0 degrees Celsius for eco.
        The comfort temperature is indicated by the sun symbol and the eco temperature by the moon symbol.

        Args:
            comfort_temperature (float): The comfort temperature in degrees Celsius.
            eco_temperature (float): The eco temperature in degrees Celsius.

        Returns:
            Status: The updated status.

        Raises:
            Eq3StateException: If the presets are not available or the thermostat is not connected.
            Eq3AlreadyAwaitingResponseException: If a status command is already pending.
            Eq3CommandException: If an error occurs while sending the command.
            Eq3TimeoutException: If the command times out.
            Eq3InvalidDataException: If the temperatures are invalid.
        """
        return await self._async_write_command_with_status_response(
            _ComfortEcoConfigureCommand(
                comfort_temperature=comfort_temperature,
                eco_temperature=eco_temperature,
            )
        )

    async def async_configure_comfort_temperature(
        self, comfort_temperature: float
    ) -> Status:
        """Set the thermostat's comfort temperature.

        Temperatures are in degrees Celsius and specified in 0.5 degree increments.
        The initial value is 21.0 degrees Celsius.
        The comfort temperature is indicated by the sun symbol.

        Args:
            comfort_temperature (float): The comfort temperature in degrees Celsius.

        Returns:
            Status: The updated status.

        Raises:
            Eq3StateException: If the presets are not available or the thermostat is not connected.
            Eq3AlreadyAwaitingResponseException: If a status command is already pending.
            Eq3CommandException: If an error occurs while sending the command.
            Eq3TimeoutException: If the command times out.
            Eq3InvalidDataException: If the temperature is invalid.
        """
        return await self.async_configure_presets(
            comfort_temperature, self.presets.eco_temperature
        )

    async def async_configure_eco_temperature(self, eco_temperature: float) -> Status:
        """Set the thermostat's eco temperature.

        Temperatures are in degrees Celsius and specified in 0.5 degree increments.
        The initial value is 17.0 degrees Celsius.
        The eco temperature is indicated by the moon symbol.

        Args:
            eco_temperature (float): The eco temperature in degrees Celsius.

        Returns:
            Status: The updated status.

        Raises:
            Eq3StateException: If the presets are not available or the thermostat is not connected.
            Eq3AlreadyAwaitingResponseException: If a status command is already pending.
            Eq3CommandException: If an error occurs while sending the command.
            Eq3TimeoutException: If the command times out.
            Eq3InvalidDataException: If the temperature is invalid.
        """
        return await self.async_configure_presets(
            self.presets.comfort_temperature, eco_temperature
        )

    async def async_configure_temperature_offset(
        self, temperature_offset: float
    ) -> Status:
        """Configure the temperature offset.

        The temperature offset is added to the measured temperature to determine the current temperature the thermostat is using internally for its calculations.
        The initial value is 0.0 degrees Celsius.
        The offset is specified in 0.5 degree increments.
        The maximum offset is EQ3_MAX_TEMP_OFFSET and the minimum offset is EQ3_MIN_TEMP_OFFSET.

        Example:
            When the thermostat is set to 20.0 degrees and the actual temperature in the room is 18.0 degrees, the offset can be set to -2.0 degrees to align the thermostat with the actual temperature.

        Args:
            temperature_offset (float): The temperature offset in degrees Celsius.

        Returns:
            Status: The updated status.

        Raises:
            Eq3StateException: If the thermostat is not connected.
            Eq3AlreadyAwaitingResponseException: If a status command is already pending.
            Eq3CommandException: If an error occurs while sending the command.
            Eq3TimeoutException: If the command times out.
            Eq3InvalidDataException: If the temperature is invalid.
        """
        return await self._async_write_command_with_status_response(
            _OffsetConfigureCommand(offset=temperature_offset)
        )

    async def async_configure_window_open(
        self, temperature: float, duration: timedelta | float | int
    ) -> Status:
        """Configure the window open behaviour.

        Temperatures are in degrees Celsius and specified in 0.5 degree increments.
        Durations are specified in 5 minute increments.
        The initial values are 12.0 degrees Celsius and 15 minutes.
        If a float is provided for the duration, it will be converted to a timedelta with minutes as the unit.

        Args:
            temperature (float): The temperature at which the window open behavior should be triggered in degrees Celsius.
            duration (timedelta | float): The duration for which the window open behavior should be active.

        Returns:
            Status: The updated status.

        Raises:
            Eq3StateException: If the presets are not available or the thermostat is not connected.
            Eq3AlreadyAwaitingResponseException: If a status command is already pending.
            Eq3CommandException: If an error occurs while sending the command.
            Eq3TimeoutException: If the command times out.
            Eq3InvalidDataException: If the temperature or duration is invalid.
        """
        if isinstance(duration, float) or isinstance(duration, int):
            duration = timedelta(minutes=duration)

        return await self._async_write_command_with_status_response(
            _WindowOpenConfigureCommand(
                window_open_temperature=temperature,
                window_open_time=duration,
            )
        )

    async def async_configure_window_open_temperature(
        self, temperature: float
    ) -> Status:
        """Configure the window open temperature.

        Temperatures are in degrees Celsius and specified in 0.5 degree increments.
        The initial value is 12.0 degrees Celsius.

        Args:
            temperature (float): The temperature at which the window open behavior should be triggered in degrees Celsius.

        Returns:
            Status: The updated status.

        Raises:
            Eq3StateException: If the presets are not available or the thermostat is not connected.
            Eq3AlreadyAwaitingResponseException: If a status command is already pending.
            Eq3CommandException: If an error occurs while sending the command.
            Eq3TimeoutException: If the command times out.
            Eq3InvalidDataException: If the temperature is invalid.
        """
        return await self.async_configure_window_open(
            temperature, self.presets.window_open_time
        )

    async def async_configure_window_open_duration(
        self,
        duration: timedelta | float,
    ) -> Status:
        """Configure the window open duration.

        The duration is specified in 5 minute increments.
        The initial value is 15 minutes.
        If a float is provided, it will be converted to a timedelta with minutes as the unit.

        Args:
            duration (timedelta | float): The duration for which the window open behavior should be active.

        Returns:
            Status: The updated status.

        Raises:
            Eq3StateException: If the presets are not available or the thermostat is not connected.
            Eq3AlreadyAwaitingResponseException: If a status command is already pending.
            Eq3CommandException: If an error occurs while sending the command.
            Eq3TimeoutException: If the command times out.
            Eq3InvalidDataException: If the duration is invalid.
        """
        return await self.async_configure_window_open(
            self.presets.window_open_temperature, duration
        )

    ### Internal ###

    async def __aenter__(self) -> Self:
        """Async context manager enter.

        Connects to the thermostat. After connecting, the device data, status, and schedule will be queried and stored.
        When the connection is established, the CONNECTED event will be triggered.

        Raises:
            Eq3StateException: If the thermostat is already connected.
            Eq3ConnectionException: If the connection fails.
            Eq3TimeoutException: If the connection times out.
            Eq3CommandException: If an error occurs while sending a command.
        """
        await self.async_connect()
        return self

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_value: BaseException | None,
        traceback: TracebackType | None,
    ) -> None:
        """Async context manager exit.

        Disconnects from the thermostat. Before disconnection all pending futures will be cancelled.
        When the disconnection is complete, the DISCONNECTED event will be triggered.

        Raises:
            Eq3StateException: If the thermostat is not connected.
            Eq3ConnectionException: If the disconnection fails.
            Eq3TimeoutException: If the disconnection times out.
        """
        await self.async_disconnect()

    async def _async_write_command_with_device_data_response(
        self, command: _Eq3Struct
    ) -> DeviceData:
        """Write a command to the thermostat and wait for a device data response.

        Args:
            command (_Eq3Struct): The command to write.

        Returns:
            DeviceData: The device data.

        Raises:
            Eq3StateException: If the thermostat is not connected.
            Eq3AlreadyAwaitingResponseException: If a device data command is already pending.
            Eq3TimeoutException: If the command times out.
            Eq3CommandException: If an error occurs while sending the command.
        """
        if self._device_data_future is not None:
            raise Eq3AlreadyAwaitingResponseException(
                "Already awaiting a device data command response"
            )

        self._device_data_future = asyncio.Future()

        await self._async_write_command(command)

        try:
            device_data = await asyncio.wait_for(
                self._device_data_future, self._command_timeout
            )
        except TimeoutError as ex:
            raise Eq3TimeoutException("Timeout during device data command") from ex
        finally:
            self._device_data_future = None

        return device_data

    async def _async_write_command_with_status_response(
        self, command: _Eq3Struct
    ) -> Status:
        """Write a command to the thermostat and wait for a status response.

        Args:
            command (_Eq3Struct): The command to write.

        Returns:
            Status: The status.

        Raises:
            Eq3StateException: If the thermostat is not connected.
            Eq3AlreadyAwaitingResponseException: If a status command is already pending.
            Eq3TimeoutException: If the command times out.
            Eq3CommandException: If an error occurs while sending the command.
        """
        if self._status_future is not None:
            raise Eq3AlreadyAwaitingResponseException(
                "Already awaiting a status command response"
            )

        self._status_future = asyncio.Future()

        await self._async_write_command(command)

        try:
            status = await asyncio.wait_for(self._status_future, self._command_timeout)
        except TimeoutError as ex:
            raise Eq3TimeoutException("Timeout during status command") from ex
        finally:
            self._status_future = None

        return status

    async def _async_write_commands_with_schedule_response(
        self, commands: list[_Eq3Struct]
    ) -> Schedule:
        """Write commands to the thermostat and wait for a schedule response.

        Args:
            commands (list[_Eq3Struct]): The commands to write.

        Returns:
            Schedule: The schedule.

        Raises:
            Eq3StateException: If the thermostat is not connected.
            Eq3AlreadyAwaitingResponseException: If a schedule command is already pending.
            Eq3TimeoutException: If the command times out.
            Eq3CommandException: If an error occurs while sending the command.
        """
        if self._schedule_future is not None:
            raise Eq3AlreadyAwaitingResponseException(
                "Already awaiting a schedule command response"
            )

        self._schedule_future = asyncio.Future()
        self._schedule_future_counter = len(commands)

        for command in commands:
            await self._async_write_command(command)

        try:
            schedule = await asyncio.wait_for(
                self._schedule_future, self._command_timeout
            )
        except TimeoutError as ex:
            raise Eq3TimeoutException("Timeout during schedule command") from ex
        finally:
            self._schedule_future = None

        return schedule

    async def _async_write_command(self, command: _Eq3Struct) -> None:
        """Write a command to the thermostat.

        Args:
            command (_Eq3Struct): The command to write.

        Raises:
            Eq3StateException: If the thermostat is not connected.
            Eq3CommandException: If an error occurs while sending the command.
            Eq3TimeoutException: If the command times out.
        """
        if not self.is_connected:
            raise Eq3StateException("Not connected")

        data = command.to_bytes()

        async with self._lock:
            try:
                await asyncio.wait_for(
                    self._conn.write_gatt_char(_Eq3Characteristic.WRITE, data),
                    self._command_timeout,
                )
            except BleakError as ex:
                raise Eq3CommandException("Error during write") from ex
            except TimeoutError as ex:
                raise Eq3TimeoutException("Timeout during write") from ex

    def _on_disconnected(self, _: BleakClient) -> None:
        """Handle disconnection from the thermostat."""
        asyncio.create_task(self._trigger_event(Eq3Event.DISCONNECTED))

    async def _on_message_received(
        self, _: BleakGATTCharacteristic, data: bytearray
    ) -> None:
        """Handle received messages from the thermostat."""
        data_bytes = bytes(data)
        command = DataclassStruct(_Eq3Message).parse(data_bytes)

        match command.cmd:
            case _Eq3Command.ID_RETURN:
                device_data = DeviceData._from_bytes(data_bytes)
                return await self._on_device_data_received(device_data)
            case _Eq3Command.INFO_RETURN:
                if not command.is_status_command:
                    return

                status = Status._from_bytes(data_bytes)
                return await self._on_status_received(status)
            case _Eq3Command.SCHEDULE_RETURN:
                schedule = Schedule._from_bytes(data_bytes)
                return await self._on_schedule_received(schedule)
            case _:
                raise Eq3InternalException(f"Unknown command: {command}")

    async def _on_device_data_received(self, device_data: DeviceData) -> None:
        """Handle received device data.

        Triggers the DEVICE_DATA_RECEIVED event.

        Args:
            device_data (DeviceData): The received device data.
        """
        self._last_device_data = device_data

        if self._device_data_future is not None:
            self._device_data_future.set_result(device_data)

        await self._trigger_event(
            Eq3Event.DEVICE_DATA_RECEIVED, device_data=device_data
        )

    async def _on_status_received(self, status: Status) -> None:
        """Handle received status.

        Triggers the STATUS_RECEIVED event.

        Args:
            status (Status): The received status.
        """
        self._last_status = status

        if self._status_future is not None:
            self._status_future.set_result(status)

        await self._trigger_event(Eq3Event.STATUS_RECEIVED, status=status)

    async def _on_schedule_received(self, schedule: Schedule) -> None:
        """Handle received schedule.

        Merges the received schedule with the last known schedule and triggers the SCHEDULE_RECEIVED event.

        Args:
            schedule (Schedule): The received schedule.
        """
        if self._last_schedule is None:
            self._last_schedule = schedule

        self._last_schedule.merge(schedule)

        if self._schedule_future is not None:
            self._schedule_future_counter -= 1

            if self._schedule_future_counter == 0:
                self._schedule_future.set_result(self._last_schedule)

        await self._trigger_event(
            Eq3Event.SCHEDULE_RECEIVED, schedule=self._last_schedule
        )

    ### Callbacks ###

    @overload
    def register_callback(
        self,
        event: Literal[Eq3Event.DISCONNECTED],
        callback: Union[Callable[[], None], Callable[[], Awaitable[None]]],
    ) -> None: ...

    @overload
    def register_callback(
        self,
        event: Union[Literal[Eq3Event.CONNECTED]],
        callback: Union[
            Callable[[DeviceData, Status, Schedule], None],
            Callable[[DeviceData, Status, Schedule], Awaitable[None]],
        ],
    ) -> None: ...

    @overload
    def register_callback(
        self,
        event: Literal[Eq3Event.DEVICE_DATA_RECEIVED],
        callback: Union[
            Callable[[DeviceData], None], Callable[[DeviceData], Awaitable[None]]
        ],
    ) -> None: ...

    @overload
    def register_callback(
        self,
        event: Literal[Eq3Event.STATUS_RECEIVED],
        callback: Union[Callable[[Status], None], Callable[[Status], Awaitable[None]]],
    ) -> None: ...

    @overload
    def register_callback(
        self,
        event: Literal[Eq3Event.SCHEDULE_RECEIVED],
        callback: Union[
            Callable[[Schedule], None], Callable[[Schedule], Awaitable[None]]
        ],
    ) -> None: ...

    def register_callback(
        self,
        event: Eq3Event,
        callback: Union[Callable[..., None], Callable[..., Awaitable[None]]],
    ) -> None:
        """Register a callback for a specific event."""
        if callback in self._callbacks[event]:
            return

        self._callbacks[event].append(callback)

    @overload
    def unregister_callback(
        self,
        event: Literal[Eq3Event.DISCONNECTED],
        callback: Union[Callable[[], None], Callable[[], Awaitable[None]]],
    ) -> None: ...

    @overload
    def unregister_callback(
        self,
        event: Union[Literal[Eq3Event.CONNECTED]],
        callback: Union[
            Callable[[DeviceData, Status, Schedule], None],
            Callable[[DeviceData, Status, Schedule], Awaitable[None]],
        ],
    ) -> None: ...

    @overload
    def unregister_callback(
        self,
        event: Literal[Eq3Event.DEVICE_DATA_RECEIVED],
        callback: Union[
            Callable[[DeviceData], None], Callable[[DeviceData], Awaitable[None]]
        ],
    ) -> None: ...

    @overload
    def unregister_callback(
        self,
        event: Literal[Eq3Event.STATUS_RECEIVED],
        callback: Union[Callable[[Status], None], Callable[[Status], Awaitable[None]]],
    ) -> None: ...

    @overload
    def unregister_callback(
        self,
        event: Literal[Eq3Event.SCHEDULE_RECEIVED],
        callback: Union[
            Callable[[Schedule], None], Callable[[Schedule], Awaitable[None]]
        ],
    ) -> None: ...

    def unregister_callback(
        self,
        event: Eq3Event,
        callback: Union[Callable[..., None], Callable[..., Awaitable[None]]],
    ) -> None:
        """Unregister a callback for a specific event."""
        if callback not in self._callbacks[event]:
            return

        self._callbacks[event].remove(callback)

    @overload
    async def _trigger_event(self, event: Literal[Eq3Event.DISCONNECTED]) -> None: ...

    @overload
    async def _trigger_event(
        self,
        event: Literal[Eq3Event.CONNECTED],
        *,
        device_data: DeviceData,
        status: Status,
        schedule: Schedule,
    ) -> None: ...

    @overload
    async def _trigger_event(
        self,
        event: Literal[Eq3Event.DEVICE_DATA_RECEIVED],
        *,
        device_data: DeviceData,
    ) -> None: ...

    @overload
    async def _trigger_event(
        self,
        event: Literal[Eq3Event.STATUS_RECEIVED],
        *,
        status: Status,
    ) -> None: ...

    @overload
    async def _trigger_event(
        self,
        event: Literal[Eq3Event.SCHEDULE_RECEIVED],
        *,
        schedule: Schedule,
    ) -> None: ...

    async def _trigger_event(
        self,
        event: Eq3Event,
        *,
        device_data: DeviceData | None = None,
        status: Status | None = None,
        schedule: Schedule | None = None,
    ) -> None:
        """Call the callbacks for a specific event."""
        async_callbacks = [
            callback
            for callback in self._callbacks[event]
            if asyncio.iscoroutinefunction(callback)
        ]
        sync_callbacks = [
            callback
            for callback in self._callbacks[event]
            if not asyncio.iscoroutinefunction(callback)
        ]

        args: (
            tuple[DeviceData, Status, Schedule]
            | tuple[DeviceData]
            | tuple[Status]
            | tuple[Schedule]
            | tuple[()]
        )

        match event:
            case Eq3Event.DISCONNECTED:
                args = ()
            case Eq3Event.CONNECTED:
                if device_data is None or status is None or schedule is None:
                    raise Eq3InternalException(
                        "device_data, status, and schedule must not be None for CONNECTED event"
                    )
                args = (device_data, status, schedule)
            case Eq3Event.DEVICE_DATA_RECEIVED:
                if device_data is None:
                    raise Eq3InternalException(
                        "device_data must not be None for DEVICE_DATA_RECEIVED event"
                    )
                args = (device_data,)
            case Eq3Event.STATUS_RECEIVED:
                if status is None:
                    raise Eq3InternalException(
                        "status must not be None for STATUS_RECEIVED event"
                    )
                args = (status,)
            case Eq3Event.SCHEDULE_RECEIVED:
                if schedule is None:
                    raise Eq3InternalException(
                        "schedule must not be None for SCHEDULE_RECEIVED event"
                    )
                args = (schedule,)

        await asyncio.gather(*[callback(*args) for callback in async_callbacks])

        for callback in sync_callbacks:
            callback(*args)



"""Constants for the eq3btsmart library."""

from enum import IntEnum, StrEnum, auto
from typing import Self

from construct_typed import EnumBase, FlagsEnumBase

__all__ = [
    "EQ3_DEFAULT_AWAY_TEMP",
    "EQ3_DEFAULT_COMFORT_TEMP",
    "EQ3_DEFAULT_ECO_TEMP",
    "EQ3_DEFAULT_WINDOW_OPEN_TEMP",
    "EQ3_DEFAULT_WINDOW_OPEN_DURATION",
    "EQ3_DEFAULT_OFFSET_TEMP",
    "EQ3_MIN_TEMP",
    "EQ3_MAX_TEMP",
    "EQ3_OFF_TEMP",
    "EQ3_ON_TEMP",
    "EQ3_MIN_OFFSET",
    "EQ3_MAX_OFFSET",
    "DEFAULT_CONNECTION_TIMEOUT",
    "DEFAULT_COMMAND_TIMEOUT",
    "Eq3WeekDay",
    "Eq3OperationMode",
    "Eq3Preset",
    "Eq3Event",
]

"""The initial away temperature in degrees Celsius."""
EQ3_DEFAULT_AWAY_TEMP = 12.0

"""The initial comfort temperature in degrees Celsius."""
EQ3_DEFAULT_COMFORT_TEMP = 21.0

"""The initial eco temperature in degrees Celsius."""
EQ3_DEFAULT_ECO_TEMP = 17.0

"""The initial window open temperature in degrees Celsius."""
EQ3_DEFAULT_WINDOW_OPEN_TEMP = 12.0

"""The initial window open duration in minutes."""
EQ3_DEFAULT_WINDOW_OPEN_DURATION = 15

"""The initial offset temperature in degrees Celsius."""
EQ3_DEFAULT_OFFSET_TEMP = 0.0

"""The minimum temperature that is still displayed as a temperature in degrees Celsius."""
EQ3_MIN_TEMP = 5.0

"""The maximum temperature that is still displayed as a temperature in degrees Celsius."""
EQ3_MAX_TEMP = 29.5

"""The off temperature in degrees Celsius."""
EQ3_OFF_TEMP = 4.5

"""The on temperature in degrees Celsius."""
EQ3_ON_TEMP = 30.0

"""The minimum offset temperature in degrees Celsius."""
EQ3_MIN_OFFSET = -3.5

"""The maximum offset temperature in degrees Celsius."""
EQ3_MAX_OFFSET = 3.5

"""The default connection timeout in seconds."""
DEFAULT_CONNECTION_TIMEOUT = 10

"""The default command timeout in seconds."""
DEFAULT_COMMAND_TIMEOUT = 5


class Eq3Event(StrEnum):
    """Event type enumeration."""

    CONNECTED = auto()
    DISCONNECTED = auto()
    DEVICE_DATA_RECEIVED = auto()
    STATUS_RECEIVED = auto()
    SCHEDULE_RECEIVED = auto()


class Eq3OperationMode(IntEnum):
    """Operation mode enumeration."""

    AUTO = 0x00
    MANUAL = 0x40
    OFF = 0x49
    ON = 0x7B
    AWAY = 0x80


class Eq3Preset(IntEnum):
    """Preset mode enumeration."""

    COMFORT = 0
    ECO = 1


class Eq3WeekDay(EnumBase):
    """Week day enumeration."""

    SATURDAY = 0
    SUNDAY = 1
    MONDAY = 2
    TUESDAY = 3
    WEDNESDAY = 4
    THURSDAY = 5
    FRIDAY = 6

    @classmethod
    def from_index(cls, index: int) -> Self:
        """Return the enum value for the given index.

        Args:
            index: The index of the week day starting with 0 for Monday.

        Returns:
            The corresponding enum value.
        """
        adjusted_index = index + 2 if index < 5 else index - 5
        return cls(adjusted_index)


class _Eq3Characteristic(StrEnum):
    """Characteristics enumeration."""

    WRITE = "3fa4585a-ce4a-3bad-db4b-b8df8179ea09"
    NOTIFY = "d0e8434d-cd29-0996-af41-6c90f4e0eb2a"


class _Eq3Command(EnumBase):
    """Command type enumeration."""

    ID_GET = 0x00
    ID_RETURN = 0x01
    INFO_RETURN = 0x02
    INFO_GET = 0x03
    SCHEDULE_SET = 0x10
    COMFORT_ECO_CONFIGURE = 0x11
    OFFSET_CONFIGURE = 0x13
    WINDOW_OPEN_CONFIGURE = 0x14
    SCHEDULE_GET = 0x20
    SCHEDULE_RETURN = 0x21
    MODE_SET = 0x40
    TEMPERATURE_SET = 0x41
    COMFORT_SET = 0x43
    ECO_SET = 0x44
    BOOST_SET = 0x45
    LOCK_SET = 0x80


class _Eq3StatusFlags(FlagsEnumBase):
    """Status flag enumeration."""

    AUTO = 0x00
    MANUAL = 0x01
    AWAY = 0x02
    BOOST = 0x04
    DST = 0x08
    WINDOW = 0x10
    LOCKED = 0x20
    UNKNOWN = 0x40
    LOW_BATTERY = 0x80



"""Python library to interact with eQ-3 Bluetooth Smart thermostats.

:copyright: (c) 2024-present Lennard Beers
:license: MIT, see LICENSE.md for more details.
"""

__title__ = "eq3btsmart"
__author__ = "Lennard Beers"
__license__ = "MIT"
__copyright__ = "Copyright (c) 2024-present"

from eq3btsmart.const import *
from eq3btsmart.exceptions import *
from eq3btsmart.models import *
from eq3btsmart.thermostat import *
